#=========================================================================
# SVStructuralTranslatorL2_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 29, 2019
"""Test the level 2 SystemVerilog structural translator."""

from __future__ import absolute_import, division, print_function

from pymtl3.datatypes import Bits1, Bits32, BitStruct
from pymtl3.dsl import Component, InPort, OutPort, Wire
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.passes.sverilog.translation.structural.SVStructuralTranslatorL2 import (
    SVStructuralTranslatorL2,
)

from .SVStructuralTranslatorL1_test import check_eq, is_sverilog_reserved


def local_do_test( m ):
  m.elaborate()
  SVStructuralTranslatorL2.is_sverilog_reserved = is_sverilog_reserved
  tr = SVStructuralTranslatorL2( m )
  tr.clear( m )
  tr.translate_structural( m )

  ports = tr.structural.decl_ports[m]
  wires = tr.structural.decl_wires[m]
  structs = tr.structural.decl_type_struct
  conns = tr.structural.connections[m]
  check_eq( ports, m._ref_ports[m] )
  check_eq( wires, m._ref_wires[m] )
  assert map(lambda x: x[0], structs) == map(lambda x: x[0], m._ref_structs)
  check_eq( map(lambda x: x[1]['def'], structs), map(lambda x: x[1], m._ref_structs) )
  check_eq( conns, m._ref_conns[m] )

def test_struct_const_structural( do_test ):
  class B( BitStruct ):
    fields = [ ( 'foo', Bits32 ) ]
    def __init__( s, foo=42 ):
      s.foo = Bits32(foo)
  class A( Component ):
    def construct( s ):
      s.in_ = B()
      s.out = OutPort( Bits32 )
      s.connect( s.out, s.in_.foo )
  a = A()
  a._ref_structs = [
    ( rdt.Struct( 'B', {'foo':rdt.Vector(32)}, ['foo'] ), \
"""\
typedef struct packed {
  logic [31:0] foo;
} B;
""" ) ]
  a._ref_ports = { a : \
"""\
  input logic [0:0] clk,
  output logic [31:0] out,
  input logic [0:0] reset\
"""
}
  a._ref_wires = { a : "" }
  a._ref_conns = { a : \
"""\
  assign out = 32'd42;\
"""
}
  # Yosys backend test reference output
  a._ref_ports_port_yosys = a._ref_ports
  a._ref_ports_wire_yosys = { a : "" }
  a._ref_ports_conn_yosys = { a : "" }
  a._ref_wires_yosys = a._ref_wires
  a._ref_conns_yosys = a._ref_conns
  # TestVectorSimulator properties
  def tv_in( m, tv ):
    pass
  def tv_out( m, tv ):
    assert m.out == Bits32(tv[0])
  a._test_vectors = [
    [       42 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  do_test( a )

def test_struct_port( do_test ):
  class B( BitStruct ):
    def __init__( s, foo=42 ):
      s.foo = Bits32(foo)
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out = OutPort( Bits32 )
      s.connect( s.out, s.in_.foo )
  a = A()
  a._ref_structs = [
    ( rdt.Struct( 'B', {'foo':rdt.Vector(32)}, ['foo'] ), \
"""\
typedef struct packed {
  logic [31:0] foo;
} B;
""" ) ]
  a._ref_ports = { a : \
"""\
  input logic [0:0] clk,
  input B in_,
  output logic [31:0] out,
  input logic [0:0] reset\
""" }
  a._ref_wires = { a : "" }
  a._ref_conns = { a : \
"""\
  assign out = in_.foo;\
""" }

  # Yosys backend test reference output
  a._ref_ports_port_yosys = { a : \
"""\
  input logic [0:0] clk,
  input logic [31:0] in_$foo,
  output logic [31:0] out,
  input logic [0:0] reset\
""" }
  a._ref_ports_wire_yosys = { a : \
"""\
  logic [31:0] in_;\
""" }
  a._ref_ports_conn_yosys = { a : \
"""\
  assign in_[31:0] = in_$foo;\
"""
}
  a._ref_wires_yosys = a._ref_wires
  a._ref_conns_yosys = { a : \
"""\
  assign out = in_$foo;\
""" }

  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_ = tv[0]
  def tv_out( m, tv ):
    assert m.out == Bits32(tv[1])
  a._test_vectors = [
    [       B(),   42 ],
    [    B( 0 ),    0 ],
    [   B( -1 ),   -1 ],
    [   B( -2 ),   -2 ],
    [   B( 24 ),   24 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  do_test( a )

def test_nested_struct_port( do_test ):
  class C( BitStruct ):
    def __init__( s, bar=1 ):
      s.bar = Bits32(bar)
  class B( BitStruct ):
    def __init__( s, foo=42 ):
      s.foo = Bits32(foo)
      s.c = C()
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out_foo = OutPort( Bits32 )
      s.out_bar = OutPort( Bits32 )
      s.connect( s.out_foo, s.in_.foo )
      s.connect( s.out_bar, s.in_.c.bar )
  a = A()
  _C = rdt.Struct( 'C', {'bar':rdt.Vector(32)}, ['bar'] )
  a._ref_structs = [ ( _C, \
"""\
typedef struct packed {
  logic [31:0] bar;
} C;
""" ),
    ( rdt.Struct( 'B', {'c':_C, 'foo':rdt.Vector(32)}, ['c', 'foo'] ), \
"""\
typedef struct packed {
  C c;
  logic [31:0] foo;
} B;
""" ) ]
  a._ref_ports = { a : \
"""\
  input logic [0:0] clk,
  input B in_,
  output logic [31:0] out_bar,
  output logic [31:0] out_foo,
  input logic [0:0] reset\
""" }
  a._ref_wires = { a : "" }
  a._ref_conns = { a : \
"""\
  assign out_foo = in_.foo;
  assign out_bar = in_.c.bar;\
""" }

  # Yosys backend test reference output
  a._ref_ports_port_yosys = { a : \
"""\
  input logic [0:0] clk,
  input logic [31:0] in_$c$bar,
  input logic [31:0] in_$foo,
  output logic [31:0] out_bar,
  output logic [31:0] out_foo,
  input logic [0:0] reset\
""" }
  a._ref_ports_wire_yosys = { a : \
"""\
  logic [31:0] in_$c;
  logic [63:0] in_;\
""" }
  a._ref_ports_conn_yosys = { a : \
"""\
  assign in_$c[31:0] = in_$c$bar;
  assign in_[63:32] = in_$c$bar;
  assign in_[31:0] = in_$foo;\
"""
}
  a._ref_wires_yosys = a._ref_wires
  a._ref_conns_yosys = { a : \
"""\
  assign out_foo = in_$foo;
  assign out_bar = in_$c$bar;\
""" }

  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_ = tv[0]
  def tv_out( m, tv ):
    assert m.out_foo == Bits32(tv[1])
    assert m.out_bar == Bits32(tv[2])
  a._test_vectors = [
    [       B(),    42,    1 ],
    [    B( 1 ),     1,    1 ],
    [   B( -1 ),    -1,    1 ],
    [   B( -2 ),    -2,    1 ],
    [   B( 24 ),    24,    1 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  do_test( a )

def test_packed_array( do_test ):
  class B( BitStruct ):
    def __init__( s, foo=42 ):
      s.foo = [ Bits32(foo) for _ in range(2) ]
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out =  [ OutPort( Bits32 ) for _ in range(2) ]
      s.connect( s.out[0], s.in_.foo[0] )
      s.connect( s.out[1], s.in_.foo[1] )
  a = A()
  _foo = rdt.PackedArray( [2], rdt.Vector(32) )
  a._ref_structs = [
    ( rdt.Struct( 'B', {'foo':_foo}, ['foo'] ), \
"""\
typedef struct packed {
  logic [1:0][31:0] foo;
} B;
""" ) ]
  a._ref_ports = { a : \
"""\
  input logic [0:0] clk,
  input B in_,
  output logic [31:0] out [0:1],
  input logic [0:0] reset\
""" }
  a._ref_wires = { a : "" }
  a._ref_conns = { a : \
"""\
  assign out[0] = in_.foo[0];
  assign out[1] = in_.foo[1];\
""" }

  # Yosys backend test reference output
  a._ref_ports_port_yosys = { a : \
"""\
  input logic [0:0] clk,
  input logic [31:0] in_$foo$__0,
  input logic [31:0] in_$foo$__1,
  output logic [31:0] out$__0,
  output logic [31:0] out$__1,
  input logic [0:0] reset\
""" }
  a._ref_ports_wire_yosys = { a : \
"""\
  logic [31:0] in_$foo [0:1];
  logic [63:0] in_;
  logic [31:0] out [0:1];\
""" }
  a._ref_ports_conn_yosys = { a : \
"""\
  assign in_$foo[0] = in_$foo$__0;
  assign in_$foo[1] = in_$foo$__1;
  assign in_[63:32] = in_$foo$__1;
  assign in_[31:0] = in_$foo$__0;
  assign out$__0 = out[0];
  assign out$__1 = out[1];\
""" }
  a._ref_wires_yosys = a._ref_wires
  a._ref_conns_yosys = { a : \
"""\
  assign out[0] = in_$foo[0];
  assign out[1] = in_$foo[1];\
""" }

  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_ = tv[0]
  def tv_out( m, tv ):
    assert m.out[0] == Bits32(tv[1])
    assert m.out[1] == Bits32(tv[2])
  a._test_vectors = [
    [       B(),    42,   42 ],
    [    B( 1 ),     1,    1 ],
    [   B( -1 ),    -1,   -1 ],
    [   B( -2 ),    -2,   -2 ],
    [   B( 24 ),    24,   24 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  do_test( a )

def test_struct_packed_array( do_test ):
  class C( BitStruct ):
    def __init__( s, bar=1 ):
      s.bar = Bits32(bar)
  class B( BitStruct ):
    def __init__( s ):
      s.c = [ C() for _ in range(2) ]
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( B )
      s.out =  [ OutPort( Bits32 ) for _ in range(2) ]
      s.connect( s.out[0], s.in_.c[0].bar )
      s.connect( s.out[1], s.in_.c[1].bar )
  a = A()
  _C = rdt.Struct('C', {'bar':rdt.Vector(32)}, ['bar'])
  a._ref_structs = [
    ( _C, \
"""\
typedef struct packed {
  logic [31:0] bar;
} C;
""" ),
    ( rdt.Struct( 'B', {'c':rdt.PackedArray([2], _C)}, ['c'] ), \
"""\
typedef struct packed {
  C [1:0] c;
} B;
""" ) ]
  a._ref_ports = { a : \
"""\
  input logic [0:0] clk,
  input B in_,
  output logic [31:0] out [0:1],
  input logic [0:0] reset\
""" }
  a._ref_wires = { a : "" }
  a._ref_conns = { a : \
"""\
  assign out[0] = in_.c[0].bar;
  assign out[1] = in_.c[1].bar;\
""" }

  # Yosys backend test reference output
  a._ref_ports_port_yosys = { a : \
"""\
  input logic [0:0] clk,
  input logic [31:0] in_$c$__0$bar,
  input logic [31:0] in_$c$__1$bar,
  output logic [31:0] out$__0,
  output logic [31:0] out$__1,
  input logic [0:0] reset\
""" }
  a._ref_ports_wire_yosys = { a : \
"""\
  logic [31:0] in_$c$bar [0:1];
  logic [31:0] in_$c [0:1];
  logic [63:0] in_;
  logic [31:0] out [0:1];\
""" }
  a._ref_ports_conn_yosys = { a : \
"""\
  assign in_$c$bar[0] = in_$c$__0$bar;
  assign in_$c[0][31:0] = in_$c$__0$bar;
  assign in_$c$bar[1] = in_$c$__1$bar;
  assign in_$c[1][31:0] = in_$c$__1$bar;
  assign in_[63:32] = in_$c$__1$bar;
  assign in_[31:0] = in_$c$__0$bar;
  assign out$__0 = out[0];
  assign out$__1 = out[1];\
""" }
  a._ref_wires_yosys = a._ref_wires
  a._ref_conns_yosys = { a : \
"""\
  assign out[0] = in_$c$bar[0];
  assign out[1] = in_$c$bar[1];\
""" }

  # TestVectorSimulator properties
  def tv_in( m, tv ):
    m.in_ = tv[0]
  def tv_out( m, tv ):
    assert m.out[0] == Bits32(tv[1])
    assert m.out[1] == Bits32(tv[2])
  a._test_vectors = [
    [       B(),     1,    1 ],
  ]
  a._tv_in, a._tv_out = tv_in, tv_out
  do_test( a )



from __future__ import absolute_import, division, print_function

from .EnqDeqIfc import DeqIfcRTL, EnqIfcRTL
from .GetGiveIfc import GetIfcRTL, GiveIfcRTL
from .ifcs_utils import enrdy_to_str, valrdy_to_str
from .MemMsg import MemMsgType, mk_mem_msg, mk_mem_req_msg, mk_mem_resp_msg
from .SendRecvIfc import RecvCL2SendRTL, RecvIfcRTL, RecvRTL2SendCL, SendIfcRTL
from .ValRdyIfc import InValRdyIfc, OutValRdyIfc
from .XcelMsg import XcelMsgType, mk_xcel_msg, mk_xcel_req_msg, mk_xcel_resp_msg



"""
========================================================================
SimpleTickPass.py
========================================================================
Generate a simple tick function (no Mamba techniques here)

Author : Shunning Jiang
Date   : Dec 26, 2018
"""
from __future__ import absolute_import, division, print_function

from graphviz import Digraph

from pymtl3.dsl.errors import UpblkCyclicError
from pymtl3.passes.BasePass import BasePass

from .errors import PassOrderError


class SimpleTickPass( BasePass ):

  def __call__( self, top ):
    if not hasattr( top._sched, "schedule" ):
      raise PassOrderError( "schedule" )

    if hasattr( top, "_cl_trace" ):
      schedule = top._cl_trace.schedule
    else:
      schedule = top._sched.schedule

    def tick_normal():
      for blk in schedule:
        blk()

    top.tick = tick_normal



#=========================================================================
# SVStructuralTranslatorL3.py
#=========================================================================
"""Provide SystemVerilog structural translator implementation."""
from __future__ import absolute_import, division, print_function

from functools import reduce

from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRType as rt
from pymtl3.passes.sverilog.util.utility import make_indent
from pymtl3.passes.translator.structural.StructuralTranslatorL3 import (
    StructuralTranslatorL3,
)

from .SVStructuralTranslatorL2 import SVStructuralTranslatorL2


class SVStructuralTranslatorL3(
    SVStructuralTranslatorL2, StructuralTranslatorL3 ):

  #-----------------------------------------------------------------------
  # Declarations
  #-----------------------------------------------------------------------

  def rtlir_tr_interface_port_decls( s, port_decls ):
    return reduce( lambda x, y: x+y, port_decls, [] )

  def rtlir_tr_interface_port_decl( s, m, port_id, port_rtype, port_array_type ):
    def _gen_ifc( id_, ifc, n_dim ):
      if not n_dim:
        ret = []
        all_properties = ifc.get_all_properties_packed()
        for name, _rtype in all_properties:
          if isinstance( _rtype, rt.Array ):
            array_type = _rtype
            rtype = _rtype.get_sub_type()
          else:
            array_type = None
            rtype = _rtype
          ret += s.rtlir_tr_interface_port_decl(
            m, id_+"$"+name, rtype, s.rtlir_tr_unpacked_array_type( array_type ) )
        return ret
      else:
        ret = []
        for i in range( n_dim[0] ):
          ret += _gen_ifc( id_+"$__"+str(i), ifc, n_dim[1:] )
        return ret
    # Port id is appended after the template of interface id:
    #   > `{id_}$msg` where `msg` is an attribute of the interface.
    if isinstance( port_rtype, rt.Port ):
      port_dtype = s.rtlir_data_type_translation( m, port_rtype.get_dtype() )
      decl_tmplt = port_rtype.get_direction() + ' ' + \
                   port_dtype['decl'] + '$' + port_id + port_array_type['decl']
      return [ decl_tmplt ]
    else:
      n_dim = port_array_type["n_dim"]
      return _gen_ifc( port_id, port_rtype, n_dim )

  def rtlir_tr_interface_decls( s, ifc_decls ):
    all_decls = reduce( lambda res, l: res + l, ifc_decls, [] )
    make_indent( all_decls, 1 )
    return ',\n'.join( all_decls )

  def rtlir_tr_interface_decl( s, ifc_id, ifc_rtype, array_type, port_decls ):
    def gen_interface_array_decl( ifc_id, ifc_rtype, n_dim, c_n_dim, port_decls ):
      # Fill in the interface id to complete the signal name:
      #   > `in_ifc$__0$msg` where `$msg` is filled when calling
      # rtlir_tr_interface_port_decl and `id_` = `in_ifc$__0`.
      ret = []
      if not n_dim:
        id_ = ifc_id + c_n_dim
        return map( lambda pdecl: pdecl.format( id_ = id_ ), port_decls )
      else:
        return reduce( lambda res, l: res+l, map(
          lambda idx: gen_interface_array_decl(
            ifc_id, ifc_rtype, n_dim[1:0], c_n_dim+'$__'+str(idx), port_decls
        ), range( n_dim[0] ) ), [] )
    n_dim = array_type['n_dim']
    return gen_interface_array_decl( ifc_id, ifc_rtype, n_dim, '', port_decls )

  #-----------------------------------------------------------------------
  # Signal operations
  #-----------------------------------------------------------------------

  def rtlir_tr_interface_array_index( s, base_signal, index ):
    return '{base_signal}$__{index}'.format( **locals() )

  def rtlir_tr_interface_attr( s, base_signal, attr ):
    return '{base_signal}${attr}'.format( **locals() )



#=========================================================================
# StructuralRTLIRGenL3Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : Apr 3, 2019
"""Provide L3 structural RTLIR generation pass."""
from __future__ import absolute_import, division, print_function

from .StructuralRTLIRGenL2Pass import StructuralRTLIRGenL2Pass
from .StructuralRTLIRSignalExpr import CurComp


class StructuralRTLIRGenL3Pass( StructuralRTLIRGenL2Pass ):
  # Override
  def contains( s, obj, signal ):
    """Return if obj contains signal.

    At L3 not all signals have direct correspondance to `s.connect`
    statements because of interfaces. Therefore we need to check if `obj`
    is some parent of `signal`.
    """
    if obj == signal:
      return True
    while not isinstance( signal, CurComp ):
      if not hasattr( signal, 'get_base' ):
        return False
      signal = signal.get_base()
      if obj == signal:
        return True
    return False



#=========================================================================
# YosysBehavioralTranslatorL5_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 9, 2019
"""Test the SystemVerilog translator implementation."""

from __future__ import absolute_import, division, print_function

from pymtl3.passes.rtlir import BehavioralRTLIRGenPass, BehavioralRTLIRTypeCheckPass
from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.passes.sverilog.translation.behavioral.test.SVBehavioralTranslatorL1_test import (
    is_sverilog_reserved,
)
from pymtl3.passes.sverilog.translation.behavioral.test.SVBehavioralTranslatorL5_test import (
    test_subcomponent,
    test_subcomponent_index,
)

from ..YosysBehavioralTranslatorL5 import YosysBehavioralRTLIRToSVVisitorL5


def local_do_test( m ):
  visitor = YosysBehavioralRTLIRToSVVisitorL5(is_sverilog_reserved)
  for comp, _all_upblks in m._ref_upblk_srcs_yosys.iteritems():
    comp.apply( BehavioralRTLIRGenPass() )
    comp.apply( BehavioralRTLIRTypeCheckPass() )
    upblks = comp._pass_behavioral_rtlir_gen.rtlir_upblks
    m_all_upblks = comp.get_update_blocks()
    for blk in m_all_upblks:
      upblk_src = visitor.enter( blk, upblks[blk] )
      upblk_src = "\n".join( upblk_src )
      assert upblk_src == _all_upblks[blk.__name__]



#=========================================================================
# BehavioralTranslatorL2_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 20, 2019
"""Test the level 2 behavioral translator."""

from __future__ import absolute_import, division, print_function

from pymtl3.datatypes import Bits16, Bits32
from pymtl3.dsl import Component, InPort, OutPort
from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.passes.translator.behavioral.BehavioralTranslatorL2 import (
    BehavioralTranslatorL2,
)

from .TestBehavioralTranslator import mk_TestBehavioralTranslator


def local_do_test( m ):
  m.elaborate()
  tr = mk_TestBehavioralTranslator(BehavioralTranslatorL2)(m)
  tr.clear( m )
  tr.translate_behavioral( m )
  upblk_src = tr.behavioral.upblk_srcs[m]
  decl_freevars = tr.behavioral.decl_freevars[m]
  decl_tmpvars = tr.behavioral.decl_tmpvars[m]
  assert upblk_src == m._ref_upblk_repr
  assert decl_freevars == m._ref_freevar_repr
  assert decl_tmpvars == m._ref_tmpvar_repr

def test_tmp_wire( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        u = s.in_ + Bits32(42)
        s.out = u
  a = A()
  a._ref_upblk_repr = \
"""\
upblk_srcs:
  upblk_src: upblk
"""
  a._ref_freevar_repr = "freevars:\n"
  a._ref_tmpvar_repr = \
"""\
tmpvars:
  tmpvar: u in upblk of Vector32
"""
  do_test( a )

def test_tmpvar_alias( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = [ OutPort( Bits32 ) for _ in range(5) ]
      @s.update
      def upblk1():
        u = s.in_ + Bits32(42)
        s.out[0] = u
      @s.update
      def upblk2():
        u = s.in_ + Bits32(42)
        s.out[1] = u
  a = A()
  a._ref_upblk_repr = \
"""\
upblk_srcs:
  upblk_src: upblk1
  upblk_src: upblk2
"""
  a._ref_freevar_repr = "freevars:\n"
  a._ref_tmpvar_repr = \
"""\
tmpvars:
  tmpvar: u in upblk1 of Vector32
  tmpvar: u in upblk2 of Vector32
"""
  do_test( a )

def test_multi_tmpvar( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = [ OutPort( Bits32 ) for _ in range(5) ]
      @s.update
      def upblk1():
        u = s.in_ + Bits32(42)
        v = s.in_ + Bits32(40)
        s.out[0] = u
        s.out[1] = u
  a = A()
  a._ref_upblk_repr = \
"""\
upblk_srcs:
  upblk_src: upblk1
"""
  a._ref_freevar_repr = "freevars:\n"
  a._ref_tmpvar_repr = \
"""\
tmpvars:
  tmpvar: u in upblk1 of Vector32
  tmpvar: v in upblk1 of Vector32
"""
  do_test( a )

def test_freevar_to_tmpvar( do_test ):
  class A( Component ):
    def construct( s ):
      STATE_IDLE = Bits32(0)
      s.out = OutPort( Bits32 )
      @s.update
      def upblk1():
        u = STATE_IDLE
        s.out = u
  a = A()
  a._ref_upblk_repr = \
"""\
upblk_srcs:
  upblk_src: upblk1
"""
  a._ref_freevar_repr = \
"""\
freevars:
  freevar: STATE_IDLE
"""
  a._ref_tmpvar_repr = \
"""\
tmpvars:
  tmpvar: u in upblk1 of Vector32
"""
  do_test( a )

def test_Bits_to_tmpvar( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits16 )
      @s.update
      def upblk1():
        u = Bits16(0)
        s.out = u
  a = A()
  a._ref_upblk_repr = \
"""\
upblk_srcs:
  upblk_src: upblk1
"""
  a._ref_freevar_repr = "freevars:\n"
  a._ref_tmpvar_repr = \
"""\
tmpvars:
  tmpvar: u in upblk1 of Vector16
"""
  do_test( a )

def test_py_int_to_tmpvar( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      @s.update
      def upblk1():
        u = 1
        s.out = u
  a = A()
  a._ref_upblk_repr = \
"""\
upblk_srcs:
  upblk_src: upblk1
"""
  a._ref_freevar_repr = "freevars:\n"
  a._ref_tmpvar_repr = \
"""\
tmpvars:
  tmpvar: u in upblk1 of Vector32
"""
  do_test( a )



from __future__ import absolute_import, division, print_function

from .arbiters import RoundRobinArbiter, RoundRobinArbiterEn
from .arithmetics import (
    Adder,
    And,
    Incrementer,
    LEComp,
    LShifter,
    LTComp,
    Mux,
    RShifter,
    Subtractor,
    ZeroComp,
)
from .Crossbar import Crossbar
from .RegisterFile import RegisterFile
from .registers import Reg, RegEn, RegEnRst, RegRst

#  from .enrdy_queues import BypassQueue1RTL, NormalQueue1RTL, PipeQueue1RTL
#  from .valrdy_queues import (
    #  BypassQueue1RTL,
    #  NormalQueue1RTL,
    #  NormalQueueRTL,
    #  PipeQueue1RTL,
#  )



#=========================================================================
# StructuralRTLIRSignalExpr.py
#=========================================================================
"""RTLIR signal expression class definitions and generation method."""

from __future__ import absolute_import, division, print_function

from functools import reduce

import pymtl3.dsl as dsl
from pymtl3.datatypes import Bits, BitStruct
from pymtl3.passes.rtlir.errors import RTLIRConversionError
from pymtl3.passes.rtlir.rtype import RTLIRDataType as rdt
from pymtl3.passes.rtlir.rtype import RTLIRType as rt


class BaseSignalExpr( object ):
  """Base abstract class of RTLIR signal expressions."""
  def __init__( s, rtype ):
    assert isinstance( rtype, rt.BaseRTLIRType ), \
      "non-RTLIR type {} encountered!".format( rtype )
    s.rtype = rtype

  def get_rtype( s ):
    return s.rtype

  def __eq__( s, other ):
    return type(s) is type(other) and s.rtype == other.rtype

  def __ne__( s, other ):
    return not s.__eq__( other )

  def __hash__( s ):
    return hash((type(s), s.rtype))

#-------------------------------------------------------------------------
# Internal signal expression classes
#-------------------------------------------------------------------------

class _Index( BaseSignalExpr ):
  """Base IR class for all index signal operations.
  
  This class implements a series of static methods that is called to tell
  if the current index signal operation is indexing on a signal of packed
  data type array or a bit selection.
  """
  def __init__( s, index_base, index, rtype ):
    super( _Index, s ).__init__( rtype )
    s.index = index
    s.base = index_base

  def __eq__( s, other ):
    return type(s) is type(other) and s.rtype == other.rtype and\
           s.index == other.index and \
           s.base == other.base

  def __hash__( s ):
    return hash((type(s), s.rtype, s.index, s.base))

  def get_index( s ):
    return s.index

  def get_base( s ):
    return s.base

  @staticmethod
  def is_packed_index( index_base, index ):
    """Return PackedIndex constructor if the given index base is a signal of packed data type."""
    base_rtype = index_base.get_rtype()
    if isinstance(base_rtype, (rt.Port, rt.Wire)) and \
       isinstance(base_rtype.get_dtype(), rdt.PackedArray):
      return PackedIndex
    else:
      return None

  @staticmethod
  def is_bit_selection( index_base, index ):
    """Return BitSelection constructor if the given index base is a vector signal."""
    base_rtype = index_base.get_rtype()
    if isinstance(base_rtype, (rt.Port, rt.Wire)) and \
       isinstance(base_rtype.get_dtype(), rdt.Vector):
      return BitSelection
    else:
      return None

class _UnpackedIndex( _Index ):
  """Base IR class for all index signal operations on unpacked arrays.
  
  Unpacked indexing operations include indexing on a port array, a wire
  array, a const array, an interface array, and a component array. This class
  implements a series of static methods to check if the given signal operation
  is one of them.
  """
  def __init__( s, index_base, index ):
    base_rtype = index_base.get_rtype()
    rtype = base_rtype.get_next_dim_type()
    super( _UnpackedIndex, s ).__init__( index_base, index, rtype )

  @staticmethod
  def is_port_index( index_base, index ):
    """Return PortIndex constructor if the given index base is a port array."""
    base_rtype = index_base.get_rtype()
    if isinstance(base_rtype, rt.Array) and \
       isinstance(base_rtype.get_sub_type(), rt.Port):
      return PortIndex
    else:
      return None

  @staticmethod
  def is_wire_index( index_base, index ):
    """Return WireIndex constructor if the given index base is a wire array."""
    base_rtype = index_base.get_rtype()
    if isinstance(base_rtype, rt.Array) and \
       isinstance(base_rtype.get_sub_type(), rt.Wire):
      return WireIndex
    else:
      return None

  @staticmethod
  def is_const_index( index_base, index ):
    """Return ConstIndex constructor if the given index base is a const array."""
    base_rtype = index_base.get_rtype()
    if isinstance(base_rtype, rt.Array) and \
       isinstance(base_rtype.get_sub_type(), rt.Const):
      return ConstIndex
    else:
      return None

  @staticmethod
  def is_ifc_view_index( index_base, index ):
    """Return InterfaceViewIndex constructor if the given index base is an interface array."""
    base_rtype = index_base.get_rtype()
    if isinstance(base_rtype, rt.Array) and \
       isinstance(base_rtype.get_sub_type(), rt.InterfaceView):
      return InterfaceViewIndex
    else:
      return None

  @staticmethod
  def is_component_index( index_base, index ):
    """Return ComponentIndex constructor if the given index base is component array."""
    base_rtype = index_base.get_rtype()
    if isinstance(base_rtype, rt.Array) and \
       isinstance(base_rtype.get_sub_type(), rt.Component):
      return ComponentIndex
    else:
      return None

class _Slice( BaseSignalExpr ):
  """Base IR class for all slicing signal operations.
  
  This class implements a single static method which checks if the given
  signal operation is slicing a vector signal (a.k.a. a part selection).
  """
  def __init__( s, slice_base, start, stop ):
    base_rtype = slice_base.get_rtype()
    dtype = base_rtype.get_dtype()
    if isinstance( base_rtype, rt.Port ):
      rtype = rt.Port( base_rtype.get_direction(), rdt.Vector( stop-start ) )
    elif isinstance( base_rtype, rt.Wire ):
      rtype = rt.Wire( rdt.Vector( stop-start ) )
    else:
      assert False, \
        "unrecognized signal type {} for slicing".format( base_rtype )
    super( _Slice, s ).__init__( rtype )
    s.base = slice_base
    s.slice = ( start, stop )

  def __eq__( s, other ):
    return type(s) is type(other) and s.rtype == other.rtype and \
           s.slice == other.slice and s.base == other.base

  def __hash__( s ):
    return hash((type(s), s.rtype, s.slice, s.base))

  def get_slice( s ):
    return s.slice

  def get_base( s ):
    return s.base

  @staticmethod
  def is_part_selection( slice_base, start, stop ):
    """Return PartSelection constructor if the given index base is a vector signal."""
    base_rtype = slice_base.get_rtype()
    if isinstance(base_rtype, (rt.Port, rt.Wire)) and \
       isinstance(base_rtype.get_dtype(), rdt.Vector) and \
       isinstance(start, int) and isinstance(stop, int):
      return PartSelection
    else:
      return None

class _Attribute( BaseSignalExpr ):
  """Base IR class for all attribute accessing signal operations.
  
  This class implements a series of static methods to check if the given signal
  operation is accessing attribute of the current component, a subcompoennt,
  an interface, or a field in a struct signal.
  """
  def __init__( s, attr_base, attr, rtype ):
    super( _Attribute, s ).__init__( rtype )
    s.attr = attr
    s.base = attr_base

  def __eq__( s, other ):
    return type(s) is type(other) and s.rtype == other.rtype and \
           s.attr == other.attr and \
           s.base == other.base

  def __hash__( s ):
    return hash((type(s), s.rtype, s.attr, s.base))

  def get_base( s ):
    return s.base

  def get_attr( s ):
    return s.attr

  @staticmethod
  def is_cur_comp_attr( attr_base, attr ):
    """Return CurCompAttr constructor if the given attribute belongs to the current component."""
    base_rtype = attr_base.get_rtype()
    if isinstance(attr_base, CurComp) and base_rtype.has_property(attr):
      return CurCompAttr
    else:
      return None

  @staticmethod
  def is_subcomp_attr( attr_base, attr ):
    """Return CurCompAttr constructor if the given attribute belongs to a sub-component."""
    base_rtype = attr_base.get_rtype()
    if not isinstance(attr_base, CurComp) and \
       isinstance(base_rtype, rt.Component) and base_rtype.has_property(attr):
      return SubCompAttr
    else:
      return None

  @staticmethod
  def is_interface_attr( attr_base, attr ):
    """Return CurCompAttr constructor if the given attribute belongs to an interface."""
    base_rtype = attr_base.get_rtype()
    if isinstance(base_rtype, rt.InterfaceView) and base_rtype.has_property(attr):
      return InterfaceAttr
    else:
      return None

  @staticmethod
  def is_struct_attr( attr_base, attr ):
    """Return CurCompAttr constructor if the given attribute belongs to struct signal."""
    base_rtype = attr_base.get_rtype()
    if isinstance(base_rtype, rt.Signal):
      dtype = base_rtype.get_dtype()
      if isinstance(dtype, rdt.Struct) and dtype.has_property(attr):
        return StructAttr
      else:
        return None
    else:
      return None

#-------------------------------------------------------------------------
# Actual IR signal operation classes
#-------------------------------------------------------------------------

class ConstInstance( BaseSignalExpr ):
  """IR class for a constant instance.
  
  This class is for constants that appear in signal operations and is not
  an attribute of a component.
  """
  def __init__( s, obj, value ):
    super( ConstInstance, s ).__init__(rt.Const(rdt.get_rtlir_dtype( obj )))
    s.value = value

  def __eq__( s, other ):
    return isinstance(other, ConstInstance) and s.rtype == other.rtype and \
           s.value == other.value

  def __hash__( s ):
    return hash((type(s), s.rtype, s.value))

  def get_value( s ):
    return s.value

class CurComp( BaseSignalExpr ):
  """IR class for the current component.
  
  This class is the base of most signal expressions because most signal
  expressions begin by accessing an attribute of the current component.
  This class implements a static method which checks if the given name
  is the same as the current component's name.
  """
  def __init__( s, comp, comp_id ):
    super( CurComp, s ).__init__(comp._pass_structural_rtlir_gen.rtlir_type)
    s.comp_id = comp_id

  def __eq__( s, other ):
    return isinstance(other, CurComp) and s.rtype == other.rtype and \
           s.comp_id == other.comp_id

  def __hash__( s ):
    return hash((type(s), s.rtype, s.comp_id))

  def get_component_id( s ):
    return s.comp_id

  @staticmethod
  def is_cur_comp( comp, comp_id ):
    """Return CurComp constructor if comp_id matches the current component's name."""
    if comp_id == comp._dsl.my_name:
      return CurComp
    else:
      return None

class PortIndex( _UnpackedIndex ):
  """IR class for indexing on a port array ."""

class WireIndex( _UnpackedIndex ):
  """IR class for indexing on a wire array ."""

class ConstIndex( _UnpackedIndex ):
  """IR class for indexing on a const array ."""

class InterfaceViewIndex( _UnpackedIndex ):
  """IR class for indexing on an interface array ."""

class ComponentIndex( _UnpackedIndex ):
  """IR class for indexing on a component array ."""

class PackedIndex( _Index ):
  """IR class for indexing on a signal of packed data type."""
  def __init__( s, index_base, index ):
    base_rtype = index_base.get_rtype()
    dtype = base_rtype.get_dtype()
    if isinstance( base_rtype, rt.Port ):
      rtype = rt.Port( base_rtype.get_direction(),
                    dtype.get_next_dim_type() )
    elif isinstance( base_rtype, rt.Wire ):
      rtype = rt.Wire( dtype.get_next_dim_type() )
    else:
      assert False, \
        "unrecognized signal type {} for indexing".format( base_rtype )
    super( PackedIndex, s ).__init__( index_base, index, rtype )

class BitSelection( _Index ):
  """IR class for selecting a bit of a vector signal."""
  def __init__( s, index_base, index ):
    base_rtype = index_base.get_rtype()
    dtype = base_rtype.get_dtype()
    if isinstance( base_rtype, rt.Port ):
      rtype = rt.Port( base_rtype.get_direction(), rdt.Vector( 1 ) )
    elif isinstance( base_rtype, rt.Wire ):
      rtype = rt.Wire( rdt.Vector( 1 ) )
    else:
      assert False, \
        "unrecognized signal type {} for indexing".format( base_rtype )
    super( BitSelection, s ).__init__( index_base, index, rtype )

class PartSelection( _Slice ):
  """IR class for selecting one or more bits of a vector signal."""

class CurCompAttr( _Attribute ):
  """IR class for accessing the attribute of the current component."""
  def __init__( s, attr_base, attr ):
    rtype = attr_base.get_rtype().get_property( attr )
    super( CurCompAttr, s ).__init__( attr_base, attr, rtype )

class SubCompAttr( _Attribute ):
  """IR class for accessing the attribute of a sub-component."""
  def __init__( s, attr_base, attr ):
    rtype = attr_base.get_rtype().get_property( attr )
    super( SubCompAttr, s ).__init__( attr_base, attr, rtype )

class InterfaceAttr( _Attribute ):
  """IR class for accessing the attribute of an interface."""
  def __init__( s, attr_base, attr ):
    rtype = attr_base.get_rtype().get_property( attr )
    super( InterfaceAttr, s ).__init__( attr_base, attr, rtype )

class StructAttr( _Attribute ):
  """IR class for accessing the attribute of a struct signal."""
  def __init__( s, attr_base, attr ):
    base_rtype = attr_base.get_rtype()
    dtype = base_rtype.get_dtype()
    if isinstance( base_rtype, rt.Port ):
      rtype = rt.Port( base_rtype.get_direction(), dtype.get_property( attr ) )
    elif isinstance( base_rtype, rt.Wire ):
      rtype = rt.Wire( dtype.get_property( attr ) )
    else:
      assert False, \
        "unrecognized signal type {} for field selection".format( base_rtype )
    super( StructAttr, s ).__init__( attr_base, attr, rtype )

#-------------------------------------------------------------------------
# Map string signal expression to IR object generation methods
#-------------------------------------------------------------------------

signal_expr_classes = {
    'Attr' :  [ _Attribute.is_cur_comp_attr,   _Attribute.is_subcomp_attr,
                _Attribute.is_interface_attr,  _Attribute.is_struct_attr ],
    'Index' : [ _Index.is_packed_index,        _Index.is_bit_selection,
                _UnpackedIndex.is_port_index,  _UnpackedIndex.is_wire_index,
                _UnpackedIndex.is_const_index, _UnpackedIndex.is_ifc_view_index,
                _UnpackedIndex.is_component_index ],
    'Slice' : [ _Slice.is_part_selection ],
    'Base'  : [ CurComp.is_cur_comp ]
}

#-------------------------------------------------------------------------
# gen_signal_expr
#-------------------------------------------------------------------------

def gen_signal_expr( cur_component, signal ):
  """Return an RTLIR signal expression for the given signal."""

  def get_next_op( expr, cur_pos, my_name, full_name ):
    """Return the next signal operation in string `expr`.
    
    Return value: ( op, *data )
    op is one of [ 'Attr', 'Slice', 'Index', 'Base', 'Done' ].
    *data is one to two variables that describes the next signal operation.
    """
    if not expr[cur_pos:]: return ( 'Done', '' ), 0
    pos = len( expr )
    dot_pos = expr.find( '.', cur_pos+1 )
    lb_pos  = expr.find( '[', cur_pos+1 )
    pos = dot_pos if dot_pos != -1 and dot_pos < pos else pos
    pos =  lb_pos if  lb_pos != -1 and  lb_pos < pos else pos

    # Attribute operation
    if expr[cur_pos] == '.': return ( 'Attr', expr[cur_pos+1:pos] ), pos

    # Index or slice operation
    elif expr[cur_pos] == '[':
      rb_pos = expr.find( ']', cur_pos )
      colon_pos = expr.find( ':', cur_pos )
      assert rb_pos != -1 and pos == rb_pos+1, \
        "unrecognized expression {}".format( expr )

      if cur_pos < colon_pos < rb_pos:
        start = int( expr[cur_pos+1:colon_pos] )
        stop  = int( expr[colon_pos+1:rb_pos] )
        return ( 'Slice', start, stop ), pos
      else:
        return ( 'Index', int(expr[cur_pos+1:rb_pos]) ), pos

    # The current component ( base of attribute )
    else:
      base_pos = expr.find( full_name )
      assert base_pos >= 0, \
        "cannot find the base of attribute {} in {}".format( full_name, expr )
      return ( 'Base', my_name ), base_pos + len( full_name )

  def get_cls_inst( func_list, cur_node, ops ):
    """Return an IR instance of the given signal operation."""
    classes = map( lambda f: f( cur_node, *ops ), func_list )
    assert reduce( lambda r, c: r + (1 if c else 0), classes, 0 ) == 1, \
      'internal error: not unique class {}!'.format( classes )
    for cls in classes:
      if cls:
        return cls( cur_node, *ops )
    assert False, \
      'internal error: no available expression nodes for {}!'.format(cur_node)

  try:

    try:
      expr = signal._dsl.full_name
      base_comp = signal
    except AttributeError:
      # Special case for a ConstInstance because it has no name
      assert hasattr( signal._dsl, 'const' ), '{} is not supported!'.format(signal)
      assert isinstance( signal._dsl.const, ( int, Bits, BitStruct ) ), \
          '{} is not an integer/BitStruct const!'.format( signal._dsl.const )
      return ConstInstance( signal, signal._dsl.const )

    # Get the base component
    base_comp = cur_component
    full_name = base_comp._dsl.full_name
    my_name = base_comp._dsl.my_name
    assert expr.find( full_name ) >= 0, \
      "cannot find the base of attribute {} in {}".format( full_name, expr )

    # Start from the base component and process one operation per iteration
    cur_pos, cur_node = 0, base_comp
    while cur_pos < len( expr ):
      op, next_pos = get_next_op( expr, cur_pos, my_name, full_name )
      if op[0] == 'Done': break
      cur_node = get_cls_inst( signal_expr_classes[ op[0] ], cur_node, op[1:] )
      cur_pos = next_pos
    return cur_node

  except AssertionError as e:
    msg = '' if e.args[0] is None else e.args[0]
    raise RTLIRConversionError( signal, msg )



#=========================================================================
# YosysBehavioralTranslatorL3_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : June 9, 2019
"""Test the SystemVerilog translator implementation."""

from __future__ import absolute_import, division, print_function

from pymtl3.passes.rtlir import BehavioralRTLIRGenPass, BehavioralRTLIRTypeCheckPass
from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.passes.sverilog.errors import SVerilogTranslationError
from pymtl3.passes.sverilog.translation.behavioral.test.SVBehavioralTranslatorL1_test import (
    is_sverilog_reserved,
)
from pymtl3.passes.sverilog.translation.behavioral.test.SVBehavioralTranslatorL3_test import (
    test_nested_struct,
    test_packed_array_behavioral,
    test_struct,
)

from ..YosysBehavioralTranslatorL3 import YosysBehavioralRTLIRToSVVisitorL3


def local_do_test( m ):
  m.elaborate()
  m.apply( BehavioralRTLIRGenPass() )
  m.apply( BehavioralRTLIRTypeCheckPass() )

  try:
    visitor = YosysBehavioralRTLIRToSVVisitorL3(is_sverilog_reserved)
    upblks = m._pass_behavioral_rtlir_gen.rtlir_upblks
    m_all_upblks = m.get_update_blocks()
    for blk in m_all_upblks:
      upblk_src = visitor.enter( blk, upblks[blk] )
      upblk_src = "\n".join( upblk_src )
      assert upblk_src == m._ref_upblk_srcs_yosys[blk.__name__]
  except SVerilogTranslationError:
    if m._ref_upblk_srcs_yosys is None:
      pass
    else:
      raise



#=========================================================================
# SVTranslator_L1_cases_test.py
#=========================================================================
"""Test the SystemVerilog translator."""

from __future__ import absolute_import, division, print_function

from pymtl3.datatypes import Bits1, Bits4, Bits32, Bits64, concat, sext, zext
from pymtl3.dsl import Component, InPort, OutPort, Wire
from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.passes.sverilog.translation.SVTranslator import SVTranslator


def trim( src ):
  lines = src.split( "\n" )
  ret = []
  for line in lines:
    if not line.startswith( "//" ):
      ret.append( line )
  return "\n".join( ret )

def local_do_test( m ):
  m.elaborate()
  tr = SVTranslator( m )
  tr.translate( m )
  assert trim( tr.hierarchy.src ) == m._ref_src

#-------------------------------------------------------------------------
# Behavioral
#-------------------------------------------------------------------------

def test_comb_assign( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = s.in_
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_,
  output logic [31:0] out,
  input logic [0:0] reset
);

  // PYMTL SOURCE:
  // 
  // @s.update
  // def upblk():
  //   s.out = s.in_
  
  always_comb begin : upblk
    out = in_;
  end

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

def test_seq_assign( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits32 )
      @s.update_on_edge
      def upblk():
        s.out = s.in_
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_,
  output logic [31:0] out,
  input logic [0:0] reset
);

  // PYMTL SOURCE:
  // 
  // @s.update_on_edge
  // def upblk():
  //   s.out = s.in_
  
  always_ff @(posedge clk) begin : upblk
    out <= in_;
  end

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

def test_concat( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_1 = InPort( Bits32 )
      s.in_2 = InPort( Bits32 )
      s.out = OutPort( Bits64 )
      @s.update
      def upblk():
        s.out = concat( s.in_1, s.in_2 )
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_1,
  input logic [31:0] in_2,
  output logic [63:0] out,
  input logic [0:0] reset
);

  // PYMTL SOURCE:
  // 
  // @s.update
  // def upblk():
  //   s.out = concat( s.in_1, s.in_2 )
  
  always_comb begin : upblk
    out = { in_1, in_2 };
  end

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

def test_concat_constants( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits64 )
      @s.update
      def upblk():
        s.out = concat( Bits32(42), Bits32(0) )
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  output logic [63:0] out,
  input logic [0:0] reset
);

  // PYMTL SOURCE:
  // 
  // @s.update
  // def upblk():
  //   s.out = concat( Bits32(42), Bits32(0) )
  
  always_comb begin : upblk
    out = { 32'd42, 32'd0 };
  end

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

def test_concat_mixed( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits64 )
      @s.update
      def upblk():
        s.out = concat( s.in_, Bits32(0) )
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_,
  output logic [63:0] out,
  input logic [0:0] reset
);

  // PYMTL SOURCE:
  // 
  // @s.update
  // def upblk():
  //   s.out = concat( s.in_, Bits32(0) )
  
  always_comb begin : upblk
    out = { in_, 32'd0 };
  end

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

def test_sext( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits64 )
      @s.update
      def upblk():
        s.out = sext( s.in_, 64 )
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_,
  output logic [63:0] out,
  input logic [0:0] reset
);

  // PYMTL SOURCE:
  // 
  // @s.update
  // def upblk():
  //   s.out = sext( s.in_, 64 )
  
  always_comb begin : upblk
    out = { { 32 { in_[31] } }, in_ };
  end

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

def test_zext( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits64 )
      @s.update
      def upblk():
        s.out = zext( s.in_, 64 )
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_,
  output logic [63:0] out,
  input logic [0:0] reset
);

  // PYMTL SOURCE:
  // 
  // @s.update
  // def upblk():
  //   s.out = zext( s.in_, 64 )
  
  always_comb begin : upblk
    out = { { 32 { 1'b0 } }, in_ };
  end

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

def test_freevar( do_test ):
  class A( Component ):
    def construct( s ):
      STATE_IDLE = Bits32(42)
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits64 )
      @s.update
      def upblk():
        s.out = concat( s.in_, STATE_IDLE )
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_,
  output logic [63:0] out,
  input logic [0:0] reset
);
  localparam [31:0] __const$STATE_IDLE = 32'd42;

  // PYMTL SOURCE:
  // 
  // @s.update
  // def upblk():
  //   s.out = concat( s.in_, STATE_IDLE )
  
  always_comb begin : upblk
    out = { in_, __const$STATE_IDLE };
  end

endmodule
"""
  a._ref_src_yosys = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_,
  output logic [63:0] out,
  input logic [0:0] reset
);

  // PYMTL SOURCE:
  // 
  // @s.update
  // def upblk():
  //   s.out = concat( s.in_, STATE_IDLE )
  
  always_comb begin : upblk
    out = { in_, 32'd42 };
  end

endmodule
"""
  do_test( a )

def test_unpacked_signal_index( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( Bits32 ) for _ in range(2) ]
      s.out = OutPort( Bits64 )
      @s.update
      def upblk():
        s.out = concat( s.in_[0], s.in_[1] )
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_ [0:1],
  output logic [63:0] out,
  input logic [0:0] reset
);

  // PYMTL SOURCE:
  // 
  // @s.update
  // def upblk():
  //   s.out = concat( s.in_[0], s.in_[1] )
  
  always_comb begin : upblk
    out = { in_[0], in_[1] };
  end

endmodule
"""
  a._ref_src_yosys = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_$__0,
  input logic [31:0] in_$__1,
  output logic [63:0] out,
  input logic [0:0] reset
);
  logic [31:0] in_ [0:1];

  // PYMTL SOURCE:
  // 
  // @s.update
  // def upblk():
  //   s.out = concat( s.in_[0], s.in_[1] )
  
  always_comb begin : upblk
    out = { in_[0], in_[1] };
  end

  assign in_[0] = in_$__0;
  assign in_[1] = in_$__1;

endmodule
"""
  do_test( a )

def test_bit_selection( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits1 )
      @s.update
      def upblk():
        s.out = s.in_[1]
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_,
  output logic [0:0] out,
  input logic [0:0] reset
);

  // PYMTL SOURCE:
  // 
  // @s.update
  // def upblk():
  //   s.out = s.in_[1]
  
  always_comb begin : upblk
    out = in_[1];
  end

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

def test_part_selection( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits64 )
      s.out = OutPort( Bits32 )
      @s.update
      def upblk():
        s.out = s.in_[4:36]
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [63:0] in_,
  output logic [31:0] out,
  input logic [0:0] reset
);

  // PYMTL SOURCE:
  // 
  // @s.update
  // def upblk():
  //   s.out = s.in_[4:36]
  
  always_comb begin : upblk
    out = in_[35:4];
  end

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

#-------------------------------------------------------------------------
# Structural
#-------------------------------------------------------------------------

def test_port_wire( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.wire_ = Wire( Bits32 )
      s.out = OutPort( Bits32 )
      s.connect( s.in_, s.wire_ )
      s.connect( s.wire_, s.out )
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_,
  output logic [31:0] out,
  input logic [0:0] reset
);
  logic [31:0] wire_;

  assign wire_ = in_;
  assign out = wire_;

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

def test_connect_constant( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = OutPort( Bits32 )
      s.connect( 42, s.out )
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  output logic [31:0] out,
  input logic [0:0] reset
);

  assign out = 32'd42;

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

def test_port_const_unaccessed( do_test ):
  class A( Component ):
    def construct( s ):
      s.STATE_IDLE = 42
      s.out = OutPort( Bits32 )
      s.connect( s.STATE_IDLE, s.out )
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  output logic [31:0] out,
  input logic [0:0] reset
);

  assign out = 32'd42;

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

def test_port_const_accessed( do_test ):
  class A( Component ):
    def construct( s ):
      s.STATE_IDLE = 42
      s.out_1 = OutPort( Bits32 )
      s.out_2 = OutPort( Bits32 )
      s.connect( s.STATE_IDLE, s.out_1 )
      @s.update
      def upblk():
        s.out_2 = s.STATE_IDLE
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  output logic [31:0] out_1,
  output logic [31:0] out_2,
  input logic [0:0] reset
);
  localparam [31:0] STATE_IDLE = 32'd42;

  // PYMTL SOURCE:
  // 
  // @s.update
  // def upblk():
  //   s.out_2 = s.STATE_IDLE
  
  always_comb begin : upblk
    out_2 = STATE_IDLE;
  end

  assign out_1 = 32'd42;

endmodule
"""
  a._ref_src_yosys = \
"""
module A
(
  input logic [0:0] clk,
  output logic [31:0] out_1,
  output logic [31:0] out_2,
  input logic [0:0] reset
);

  // PYMTL SOURCE:
  // 
  // @s.update
  // def upblk():
  //   s.out_2 = s.STATE_IDLE
  
  always_comb begin : upblk
    out_2 = 32'd42;
  end

  assign out_1 = 32'd42;

endmodule
"""
  do_test( a )

def test_port_const_array( do_test ):
  class A( Component ):
    def construct( s ):
      s.STATES = [ 1, 2, 3, 4, 5 ]
      s.out = [ OutPort( Bits32 ) for _ in range(5) ]
      s.tmp = OutPort( Bits32 )
      for i in range(5):
        s.connect( s.STATES[i], s.out[i] )
      @s.update
      def upblk():
        s.tmp = s.STATES[0]
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  output logic [31:0] out [0:4],
  input logic [0:0] reset,
  output logic [31:0] tmp
);
  localparam [31:0] STATES [0:4] = '{ 32'd1, 32'd2, 32'd3, 32'd4, 32'd5 };

  // PYMTL SOURCE:
  // 
  // @s.update
  // def upblk():
  //   s.tmp = s.STATES[0]
  
  always_comb begin : upblk
    tmp = STATES[0];
  end

  assign out[0] = 32'd1;
  assign out[1] = 32'd2;
  assign out[2] = 32'd3;
  assign out[3] = 32'd4;
  assign out[4] = 32'd5;

endmodule
"""
  a._ref_src_yosys = \
"""
module A
(
  input logic [0:0] clk,
  output logic [31:0] out$__0,
  output logic [31:0] out$__1,
  output logic [31:0] out$__2,
  output logic [31:0] out$__3,
  output logic [31:0] out$__4,
  input logic [0:0] reset,
  output logic [31:0] tmp
);
  logic [31:0] out [0:4];

  // PYMTL SOURCE:
  // 
  // @s.update
  // def upblk():
  //   s.tmp = s.STATES[0]
  
  always_comb begin : upblk
    tmp = 32'd1;
  end

  assign out$__0 = out[0];
  assign out$__1 = out[1];
  assign out$__2 = out[2];
  assign out$__3 = out[3];
  assign out$__4 = out[4];
  assign out[0] = 32'd1;
  assign out[1] = 32'd2;
  assign out[2] = 32'd3;
  assign out[3] = 32'd4;
  assign out[4] = 32'd5;

endmodule
"""
  do_test( a )

def test_port_bit_selection( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits1 )
      s.connect( s.out, s.in_[2] )
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_,
  output logic [0:0] out,
  input logic [0:0] reset
);

  assign out = in_[2:2];

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

def test_port_part_selection( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
      s.out = OutPort( Bits4 )
      s.connect( s.out, s.in_[2:6] )
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  input logic [31:0] in_,
  output logic [3:0] out,
  input logic [0:0] reset
);

  assign out = in_[5:2];

endmodule
"""
  a._ref_src_yosys = a._ref_src
  do_test( a )

def test_port_wire_array_index( do_test ):
  class A( Component ):
    def construct( s ):
      s.out = [ OutPort( Bits32 ) for _ in range(5) ]
      s.wire_ = [ Wire(Bits32) for _ in range(5) ]
      for i in range(5):
        s.connect( s.wire_[i], s.out[i] )
        s.connect( s.wire_[i], i )
  a = A()
  a._ref_src = \
"""
module A
(
  input logic [0:0] clk,
  output logic [31:0] out [0:4],
  input logic [0:0] reset
);
  logic [31:0] wire_ [0:4];

  assign out[0] = wire_[0];
  assign wire_[0] = 32'd0;
  assign out[1] = wire_[1];
  assign wire_[1] = 32'd1;
  assign out[2] = wire_[2];
  assign wire_[2] = 32'd2;
  assign out[3] = wire_[3];
  assign wire_[3] = 32'd3;
  assign out[4] = wire_[4];
  assign wire_[4] = 32'd4;

endmodule
"""
  a._ref_src_yosys = \
"""
module A
(
  input logic [0:0] clk,
  output logic [31:0] out$__0,
  output logic [31:0] out$__1,
  output logic [31:0] out$__2,
  output logic [31:0] out$__3,
  output logic [31:0] out$__4,
  input logic [0:0] reset
);
  logic [31:0] out [0:4];
  logic [31:0] wire_ [0:4];

  assign out$__0 = out[0];
  assign out$__1 = out[1];
  assign out$__2 = out[2];
  assign out$__3 = out[3];
  assign out$__4 = out[4];
  assign out[0] = wire_[0];
  assign wire_[0] = 32'd0;
  assign out[1] = wire_[1];
  assign wire_[1] = 32'd1;
  assign out[2] = wire_[2];
  assign wire_[2] = 32'd2;
  assign out[3] = wire_[3];
  assign wire_[3] = 32'd3;
  assign out[4] = wire_[4];
  assign wire_[4] = 32'd4;

endmodule
"""
  do_test( a )



from __future__ import absolute_import, division, print_function

from pymtl3.dsl import Component

from .CLLineTracePass import CLLineTracePass
from .DynamicSchedulePass import DynamicSchedulePass
from .GenDAGPass import GenDAGPass
from .LineTraceParamPass import LineTraceParamPass
from .mamba.HeuristicTopoPass import HeuristicTopoPass
from .mamba.TraceBreakingSchedTickPass import TraceBreakingSchedTickPass
from .mamba.UnrollTickPass import UnrollTickPass
from .OpenLoopCLPass import OpenLoopCLPass
from .SimpleSchedulePass import SimpleSchedulePass
from .SimpleTickPass import SimpleTickPass
from .VcdGenerationPass import VcdGenerationPass
from .WrapGreenletPass import WrapGreenletPass

SimpleSim = [
  Component.elaborate,
  GenDAGPass(),
  WrapGreenletPass(),
  SimpleSchedulePass(),
  CLLineTracePass(),
  # VcdGenerationPass(),
  SimpleTickPass(),
  LineTraceParamPass(),
  Component.lock_in_simulation
]

DynamicSim = [
  Component.elaborate,
  GenDAGPass(),
  WrapGreenletPass(),
  DynamicSchedulePass(),
  CLLineTracePass(),
  SimpleTickPass(),
  Component.lock_in_simulation
]

# This pass is created to be used for 2019 isca tutorial.
SimulationPass = [
  GenDAGPass(),
  WrapGreenletPass(),
  DynamicSchedulePass(),
  CLLineTracePass(),
  VcdGenerationPass(),
  SimpleTickPass(),
  Component.lock_in_simulation
]

OpenLoopCLSim = [
  Component.elaborate,
  GenDAGPass(),
  OpenLoopCLPass(), # Inject this pass to build infrastructure
  SimpleSchedulePass(), # Still generate schedule and tick
  # TODO: make OpenLoopCLPass work with CLLineTracePass
  SimpleTickPass(),
  Component.lock_in_simulation
]

UnrollSim = [
  Component.elaborate,
  GenDAGPass(),
  WrapGreenletPass(),
  SimpleSchedulePass(),
  UnrollTickPass(),
  Component.lock_in_simulation
]

HeuTopoUnrollSim = [
  Component.elaborate,
  GenDAGPass(),
  HeuristicTopoPass(),
  UnrollTickPass(),
  Component.lock_in_simulation
]

TraceBreakingSim = [
  Component.elaborate,
  GenDAGPass(),
  TraceBreakingSchedTickPass(),
  Component.lock_in_simulation
]



#=========================================================================
# RTLIRType.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 31, 2019
"""RTLIR instance types and generation methods.

This file contains the definitions of RTLIR instance types and methods
that generate RTLIR instances. Each instance of the non-abstract class
listed in this module is a type instance or simply a type in the RTLIR
type system. Signal types is parameterized by data types defined in the
RTLIR data type module.
"""
from __future__ import absolute_import, division, print_function

import copy
import inspect
from functools import reduce

import pymtl3.dsl as dsl
from pymtl3.datatypes import Bits, BitStruct

from ..errors import RTLIRConversionError
from ..util.utility import collect_objs
from . import RTLIRDataType as rdt


class BaseRTLIRType( object ):
  """Base abstract class for all RTLIR instance types."""
  def __ne__( s, other ):
    return not s.__eq__( other )

class NoneType( BaseRTLIRType ):
  """Type for not yet typed RTLIR temporary variables."""
  def __eq__( s, other ):
    return isinstance( other, NoneType )

  def __hash__( s ):
    return hash(type(s))

  def __str__( s ):
    return 'NoneType'

  def __repr__( s ):
    return 'NoneType'

class Array( BaseRTLIRType ):
  """Unpacked RTLIR array type."""
  def __init__( s, dim_sizes, sub_type, obj = None, unpacked = False ):
    assert isinstance( sub_type, BaseRTLIRType ), \
      "array subtype {} is not RTLIR type!".format( sub_type )
    assert not isinstance( sub_type, Array ), \
      "array subtype {} should not be array RTLIR type!".format( sub_type )
    assert len( dim_sizes ) >= 1, "array should be non-empty!"
    assert reduce( lambda s, i: s+i, dim_sizes, 0 ) > 0, \
      "array should have at least one element!"
    s.dim_sizes = dim_sizes
    s.sub_type = sub_type
    s.unpacked = unpacked
    s.obj = obj

  def _is_unpacked( s ):
    return s.unpacked

  def __eq__( s, other ):
    if not isinstance( other, Array ): return False
    if s.dim_sizes != other.dim_sizes: return False
    return s.sub_type == other.sub_type

  def __hash__( s ):
    return hash((type(s), tuple(s.dim_sizes), s.sub_type))

  def get_obj( s ):
    return s.obj

  def get_next_dim_type( s ):
    if len( s.dim_sizes ) == 1: return copy.copy( s.sub_type )
    _s = copy.copy( s )
    _s.dim_sizes = s.dim_sizes[1:]
    return _s

  def get_dim_sizes( s ):
    return s.dim_sizes

  def get_sub_type( s ):
    return s.sub_type

  def __call__( s, obj ):
    """Return if obj be cast into type `s`."""
    return s == obj

  def __str__( s ):
    return 'Array'

  def __repr__( s ):
    return 'Array{} of {}'.format( s.dim_sizes, s.sub_type )

class Signal( BaseRTLIRType ):
  """Signal abstract RTLIR instance type.

  A Signal can be a Port, a Wire, or a Const.
  """
  def __init__( s, dtype, unpacked = False ):
    assert isinstance( dtype, rdt.BaseRTLIRDataType ), \
      "signal parameterized by non-RTLIR data type {}!".format( dtype )
    s.dtype = dtype
    s.unpacked = unpacked

  def __hash__( s ):
    return hash((type(s), s.dtype))

  def is_packed_indexable( s ):
    return isinstance( s.dtype, rdt.PackedArray )

  def get_dtype( s ):
    return s.dtype

  def _is_unpacked( s ):
    return s.unpacked

class Port( Signal ):
  """Port RTLIR instance type."""
  def __init__( s, direction, dtype, unpacked = False ):
    super( Port, s ).__init__( dtype, unpacked )
    s.direction = direction

  def __eq__( s, other ):
    return isinstance(other, Port) and s.dtype == other.dtype and \
           s.direction == other.direction

  def __hash__( s ):
    return hash((type(s), s.dtype, s.direction))

  def __str__( s ):
    return 'Port'

  def __repr__( s ):
    return 'Port of {}'.format( s.dtype )

  def get_direction( s ):
    return s.direction

  def get_next_dim_type( s ):
    assert s.is_packed_indexable(), "cannot index on unindexable port!"
    return Port( s.direction, s.dtype.get_next_dim_type(), s.unpacked )

class Wire( Signal ):
  """Wire RTLIR instance type."""
  def __init__( s, dtype, unpacked = False ):
    super( Wire, s ).__init__( dtype, unpacked )

  def __eq__( s, other ):
    return isinstance(other, Wire) and s.dtype == other.dtype

  def __hash__( s ):
    return hash((type(s), s.dtype))

  def __str__( s ):
    return 'Wire'

  def __repr__( s ):
    return 'Wire of {}'.format( s.dtype )

  def get_next_dim_type( s ):
    assert s.is_packed_indexable(), "cannot index on unindexable wire!"
    return Wire( s.dtype.get_next_dim_type(), s.unpacked )

class NetWire( Signal ):
  """NetWire RTLIR instance type.
  
  Objects of this class are generated by the behavioral RTLIR type check pass
  to mark an expression as not available for any signal operations ( i.e.
  bit selection, part selection, attribute access ). This is useful because
  some backend tools ( like verilator ) does not support indexing/slicing over
  an already indexed/sliced signal ( the "net wire" ).
  """
  def __init__( s, dtype, unpacked = False ):
    super( NetWire, s ).__init__( dtype, unpacked )

  def __eq__( s, other ):
    return isinstance(other, NetWire) and s.dtype == other.dtype

  def __hash__( s ):
    return hash((type(s), s.dtype))

  def __str__( s ):
    return 'NetWire'

  def __repr__( s ):
    return 'NetWire of {}'.format( s.dtype )

  def get_next_dim_type( s ):
    assert s.is_packed_indexable(), "cannot index on unindexable net wire!"
    return NetWire( s.dtype.get_next_dim_type(), s.unpacked )

class Const( Signal ):
  """Const RTLIR instance type."""
  def __init__( s, dtype, obj = None, unpacked = False ):
    super( Const, s ).__init__( dtype, unpacked )
    s.obj = obj

  def __eq__( s, other ):
    return isinstance(other, Const) and s.dtype == other.dtype

  def __hash__( s ):
    return hash((type(s), s.dtype))

  def __str__( s ):
    return 'Const'

  def __repr__( s ):
    return 'Const of {}'.format( s.dtype )

  def get_object( s ):
    return s.obj

  def get_next_dim_type( s ):
    assert s.is_packed_indexable(), "cannot index on unindexable constant!"
    return Const( s.dtype.get_next_dim_type(), s.unpacked )

class InterfaceView( BaseRTLIRType ):
  """RTLIR instance type for a view of an interface."""
  def __init__( s, name, properties, obj = None, unpacked = False ):
    s.name = name
    s.properties = properties
    s.obj = obj
    s.cls = obj.__class__
    s.unpacked = unpacked

    if obj is not None:
      try:
        s.args = obj._dsl.args
        s.kwargs = obj._dsl.kwargs
      except AttributeError:
        assert False, "interface object {} is not constructed!".format( s.obj )

    # Sanity check
    for name, rtype in properties.iteritems():
      assert isinstance(name, str) and _is_of_type(rtype, (Port, InterfaceView)), \
        "invalid attribute {} of interface {}: only ports and interfaces allowed!". \
          format( name, s.name )

  def __str__( s ):
    return 'InterfaceView ' + s.name

  def __repr__( s ):
    return 'InterfaceView {}'.format( s.name )

  def _is_unpacked( s ):
    return s.unpacked

  def __eq__( s, other ):
    return isinstance(other, InterfaceView) and s.name == other.name

  def __hash__( s ):
    return hash((type(s), s.name))

  def get_name( s ):
    return s.name

  def get_class( s ):
    return s.cls

  def get_args( s ):
    if s.obj is not None:
      return ( s.args, s.kwargs )
    else:
      return ( [], {} )

  def get_input_ports( s ):
    return sorted(filter(
      lambda id__port: id__port[1].direction == 'input',
      s.properties.iteritems()
    ), key = lambda kv: kv[0])

  def get_output_ports( s ):
    return sorted(filter(
      lambda id__port1: id__port1[1].direction == 'output',
      s.properties.iteritems()
    ), key = lambda kv: kv[0])

  def has_property( s, p ):
    return p in s.properties

  def get_property( s, p ):
    return s.properties[ p ]

  def get_all_ports( s ):
    return sorted(filter(
      lambda name_port: isinstance( name_port[1], Port ),
      s.properties.iteritems()
    ), key = lambda kv: kv[0])

  def get_all_ports_packed( s ):
    return sorted(filter(
      lambda id__t2: \
        ( isinstance( id__t2[1], Port ) and not id__t2[1]._is_unpacked() ) or \
        ( isinstance( id__t2[1], Array ) and isinstance( id__t2[1].get_sub_type(), Port ) \
          and not id__t2[1]._is_unpacked() ),
      s.properties.iteritems()
    ), key = lambda kv: kv[0])

  def get_all_properties( s ):
    return sorted( s.properties.iteritems(), lambda kv: kv[0] )

  def get_all_properties_packed( s ):
    return sorted( filter( lambda x: not x[1]._is_unpacked(), s.properties.iteritems() ),
        key = lambda kv: kv[0] )

class Component( BaseRTLIRType ):
  """RTLIR instance type for a component."""
  def __init__( s, obj, properties, unpacked = False ):
    s.name = obj.__class__.__name__
    s.params = s._gen_parameters( obj )
    s.properties = properties
    s.unpacked = unpacked
    cls = obj.__class__
    try:
      file_name = inspect.getsourcefile( cls )
      line_no = inspect.getsourcelines( cls )[1]
      s.file_info = "File: {file_name}, Line: {line_no}".format( **locals() )
    except IOError:
      s.file_info = "Dynamically generated component " + cls.__name__

  def _gen_parameters( s, obj ):
    # s.argspec: static code reflection results
    # _dsl.args: all unnamed arguments supplied to construct()
    # _dsl.kwargs: all named arguments supplied to construct()
    try:
      argspec = inspect.getfullargspec( getattr( obj, 'construct' ) )
      assert not argspec.varkw, "keyword args are not allowed for construct!"
      assert not argspec.kwonlyargs, "keyword args are not allowed for construct!"
    except AttributeError:
      argspec = inspect.getargspec( getattr( obj, 'construct' ) )
      assert not argspec.keywords, "keyword args are not allowed for construct!"
    assert not argspec.varargs, "varargs are not allowed for construct!"
    arg_names = argspec.args[1:]

    defaults = argspec.defaults or ()
    num_args = len(arg_names)
    num_supplied = len(obj._dsl.args) + len(obj._dsl.kwargs)
    num_defaults = len(defaults)

    # No default values: each arg is either keyword or unnamed
    # Has default values: num. supplied values + num. of defaults >= num. args
    assert num_args == num_supplied or num_args <= num_supplied + num_defaults, \
        "internal error: fail to parse the arguments!"
    use_defaults = num_args != num_supplied

    ret = []
    # Handle method construct arguments
    for idx, arg_name in enumerate(arg_names):

      # Use values from _dsl.args
      if idx < len(obj._dsl.args):
        ret.append((arg_name, obj._dsl.args[idx]))

      # Use values from _dsl.kwargs
      elif arg_name in obj._dsl.kwargs:
        ret.append((arg_name, obj._dsl.kwargs[arg_name]))

      # Use default values
      else:
        assert use_defaults, "internal error: didn't expect to use default values!"
        ret.append((arg_name, defaults[idx-len(arg_names)]))

    return ret

  def _is_unpacked( s ):
    return s.unpacked

  def __eq__( s, other ):
    return isinstance(other, Component) and s.name == other.name and \
           s.params == other.params

  def __hash__( s ):
    return hash((type(s), s.name, tuple(s.params)))

  def __str__( s ):
    return 'Component'

  def __repr__( s ):
    return 'Component {}'.format( s.name )

  def get_name( s ):
    return s.name

  def get_file_info( s ):
    return s.file_info

  def get_params( s ):
    return s.params

  def get_ports( s ):
    return sorted(filter(
      lambda id__port3: isinstance( id__port3[1], Port ),
      s.properties.iteritems()
    ), key = lambda kv: kv[0])

  def get_ports_packed( s ):
    return sorted(filter(
      lambda id__t4: \
        ( isinstance( id__t4[1], Port ) and not id__t4[1]._is_unpacked() ) or \
        ( isinstance( id__t4[1], Array ) and isinstance( id__t4[1].get_sub_type(), Port ) \
          and not id__t4[1]._is_unpacked() ),
      s.properties.iteritems()
    ), key = lambda kv: kv[0])

  def get_wires( s ):
    return sorted(filter(
      lambda id__wire: isinstance( id__wire[1], Wire ),
      s.properties.iteritems()
    ), key = lambda kv: kv[0])

  def get_wires_packed( s ):
    return sorted(filter(
      lambda id__t5: \
        ( isinstance( id__t5[1], Wire ) and not id__t5[1]._is_unpacked() ) or \
        ( isinstance( id__t5[1], Array ) and isinstance( id__t5[1].get_sub_type(), Wire ) \
          and not id__t5[1]._is_unpacked() ),
      s.properties.iteritems()
    ), key = lambda kv: kv[0])

  def get_consts( s ):
    return sorted(filter(
      lambda id__const: isinstance( id__const[1], Const ),
      s.properties.iteritems()
    ), key = lambda kv: kv[0])

  def get_consts_packed( s ):
    return sorted(filter(
      lambda id__t6: \
        ( isinstance( id__t6[1], Const ) and not id__t6[1]._is_unpacked() ) or \
        ( isinstance( id__t6[1], Array ) and isinstance( id__t6[1].get_sub_type(), Const ) \
          and not id__t6[1]._is_unpacked() ),
      s.properties.iteritems()
    ), key = lambda kv: kv[0])

  def get_ifc_views( s ):
    return sorted(filter(
      lambda id__ifc: isinstance( id__ifc[1], InterfaceView ),
      s.properties.iteritems()
    ), key = lambda kv: kv[0])

  def get_ifc_views_packed( s ):
    return sorted(filter(
      lambda id__t7: \
        ( isinstance( id__t7[1], InterfaceView ) and not id__t7[1]._is_unpacked() ) or \
        ( isinstance( id__t7[1], Array ) and isinstance( id__t7[1].get_sub_type(),InterfaceView ) \
          and not id__t7[1]._is_unpacked() ),
      s.properties.iteritems()
    ), key = lambda kv: kv[0])

  def get_subcomps( s ):
    return sorted(filter(
      lambda id__subcomp: isinstance( id__subcomp[1], Component ),
      s.properties.iteritems()
    ), key = lambda kv: kv[0])

  def get_subcomps_packed( s ):
    return sorted(filter(
      lambda id__t9: \
        ( isinstance( id__t9[1], Component ) and not id__t9[1]._is_unpacked() ) or \
        ( isinstance( id__t9[1], Array ) and isinstance( id__t9[1].get_sub_type(), Component ) \
          and not id__t9[1]._is_unpacked() ),
      s.properties.iteritems()
    ), key = lambda kv: kv[0])

  def has_property( s, p ):
    return p in s.properties

  def get_property( s, p ):
    return s.properties[ p ]

  def get_all_properties( s ):
    return s.properties

#-------------------------------------------------------------------------
# Internal Methods
#-------------------------------------------------------------------------

def _is_rtlir_ifc_convertible( obj ):
  valid_ifc_attributes = ( dsl.InPort, dsl.OutPort, dsl.Interface )
  if isinstance( obj, list ):
    while isinstance( obj, list ):
      # Empty lists will be dropped
      if len( obj ) == 0:
        return True
      # assert len( obj ) > 0, "one dimension of {} is 0!".format( obj )
      obj = obj[0]
    return _is_rtlir_ifc_convertible( obj )
  elif isinstance( obj, valid_ifc_attributes ):
    return True
  else:
    return False

def _freeze( obj ):
  if isinstance( obj, list ):
    return tuple( _freeze(value) for value in obj )
  else:
    return obj

def _unpack( id_, Type ):
  if not isinstance( Type, Array ): return [ ( id_, Type ) ]
  ret = []
  for idx in range( Type.get_dim_sizes()[0] ):
    ret.append( ( id_+'[{}]'.format(idx), Type.get_next_dim_type() ) )
    ret.extend(_unpack(id_+'[{}]'.format( idx ), Type.get_next_dim_type()))
  return ret

def _add_packed_instances( id_, Type, properties ):
  assert isinstance( Type, Array ), "{} is not an array type!".format( Type )
  for _id, _Type in _unpack( id_, Type ):
    assert hasattr( _Type, 'unpacked' ), \
      "{} {} is not unpacked!".format( _Type, _id )
    _Type.unpacked = True
    properties[ _id ] = _Type

def _handle_Array( _id, _obj ):
  obj = _obj
  if len( obj ) == 0:
    return None
  ref_type = get_rtlir( obj[0] )
  assert \
    reduce( lambda res,i: res and (get_rtlir(i)==ref_type), obj, True ), \
    'all elements of array {} must have the same type {}!'.format(
      obj, ref_type )
  dim_sizes = []
  while isinstance( obj, list ):
    if len( obj ) == 0:
      return None
    dim_sizes.append( len( obj ) )
    obj = obj[0]
  if isinstance( obj, ( int, Bits ) ):
    return Array( dim_sizes, get_rtlir( obj ), _obj )
  else:
    return Array( dim_sizes, get_rtlir( obj ) )

def _handle_InPort( p_id, obj ):
  return Port( 'input', rdt.get_rtlir_dtype( obj ) )

def _handle_OutPort( p_id, obj ):
  return Port( 'output', rdt.get_rtlir_dtype( obj ) )

def _handle_Wire( w_id, obj ):
  return Wire( rdt.get_rtlir_dtype( obj ) )

def _handle_Const( c_id, obj ):
  return Const( rdt.get_rtlir_dtype( obj ), obj )

def _handle_Interface( i_id, obj ):
  properties = {}
  collected_objs = collect_objs( obj, object )
  for _id, _obj in collected_objs:
    if _is_rtlir_ifc_convertible( _obj ):
      _obj_type = get_rtlir( _obj )
      if _obj_type is not None:
        properties[ _id ] = _obj_type
        if isinstance( _obj_type, Array ):
          _add_packed_instances( _id, _obj_type, properties )
    # TODO: Figure out a way to inform user of dropped attributes without
    # flooding STDOUT
    # else:
      # err_msg = \
# """\
 # - Note: {} attribute {} of {} was dropped during conversion to RTLIR because it is
         # not an interface, a port, or a list of them. \
# """
      # print( err_msg.format( _id, _obj, i_id ) )
  return InterfaceView( obj.__class__.__name__, properties, obj )

def _handle_Component( c_id, obj ):
  properties = {}
  collected_objs = collect_objs( obj, object )
  for _id, _obj in collected_objs:
    # Untranslatable attributes will be ignored
    if is_rtlir_convertible( _obj ):
      _obj_type = get_rtlir( _obj )
      if _obj_type is not None:
        properties[ _id ] = _obj_type
        if isinstance( _obj_type, Array ):
          _add_packed_instances( _id, _obj_type, properties )
    # TODO: Figure out a way to inform user of dropped attributes without
    # flooding STDOUT
    # else:
      # err_msg = \
# """\
 # - Note: {} attribute {} of {} was dropped during conversion to RTLIR because it is
         # not a port, a, wire, an interface, a component, a constantor, or a
         # list of them. \
# """
      # print( err_msg.format( _id, _obj, c_id ) )
  return Component( obj, properties )

def _is_of_type( obj, Type ):
  """Return True is `obj` is of RTLIR type `Type`."""
  if isinstance( obj, Type ):
    return True
  if isinstance( obj, Array ) and isinstance( obj.get_sub_type(), Type ):
    return True
  return False

_RTLIR_ifc_handlers = [
  ( list,          _handle_Array ),
  ( dsl.InPort,    _handle_InPort ),
  ( dsl.OutPort,   _handle_OutPort ),
  ( dsl.Interface, _handle_Interface ),
]

_RTLIR_handlers = [
  ( list,                     _handle_Array ),
  ( dsl.InPort,               _handle_InPort ),
  ( dsl.OutPort,              _handle_OutPort ),
  ( dsl.Wire,                 _handle_Wire ),
  ( ( int, Bits, BitStruct ), _handle_Const ),
  ( dsl.Interface,            _handle_Interface ),
  ( dsl.Component,            _handle_Component ),
]

#-------------------------------------------------------------------------
# Public APIs
#-------------------------------------------------------------------------

def get_component_ifc_rtlir( obj ):
  """Return the RTLIR of the interfaces of component `obj`."""
  primitive_types = tuple( x[0] for x in _RTLIR_ifc_handlers[1:] )

  def _is_interface( id_, obj ):
    _type = type(obj)
    if isinstance( obj, primitive_types ):
      return True
    if not isinstance( obj, list ):
      return False
    while isinstance( obj, list ):
      assert len( obj ) > 0, "{} is an empty list!".format( id_ )
      obj = obj[0]
    return isinstance( obj, primitive_types )

  def _get_ifc_rtlir( id_, obj ):
    for Type, handler in _RTLIR_ifc_handlers:
      if isinstance( obj, Type ):
        return handler( id_, obj )
    return None

  try:
    assert isinstance(obj, dsl.Component), \
      "the given object is not a PyMTL component!"
    properties = {}
    collected_objs = collect_objs( obj, object )
    for _id, _obj in collected_objs:
      if _is_interface( _id, _obj ):
        _obj_type = _get_ifc_rtlir( _id, _obj )
        if _obj_type is not None:
          properties[ _id ] = _obj_type
          if isinstance( _obj_type, Array ):
            _add_packed_instances( _id, _obj_type, properties )
    return Component( obj, properties )
  except AssertionError as e:
    msg = '' if e.args[0] is None else e.args[0]
    raise RTLIRConversionError( obj, msg )

def is_rtlir_convertible( obj ):
  """Return if `obj` can be converted into an RTLIR instance."""
  pymtl_constructs = (
    dsl.InPort, dsl.OutPort, dsl.Wire,
    Bits, BitStruct, dsl.Interface, dsl.Component,
  )
  # TODO: improve this long list of isinstance check
  if isinstance( obj, list ):
    while isinstance( obj, list ):
      # Empty lists will be dropped
      if len( obj ) == 0:
        return True
      obj = obj[0]
    return is_rtlir_convertible( obj )
  elif isinstance( obj, pymtl_constructs ):
    return True
  elif isinstance( obj, int ):
    return True
  else:
    return False

__rtlir_cache = {}

def get_rtlir( _obj ):
  """Return an RTLIR instance corresponding to `obj`."""
  obj = _freeze( _obj )
  if obj in __rtlir_cache:
    return __rtlir_cache[ obj ]
  else:
    try:
      for Type, handler in _RTLIR_handlers:
        if isinstance( _obj, Type ):
          __rtlir_cache[ obj ] = handler( "<name not available>", _obj )
          return __rtlir_cache[ obj ]
      # Cannot convert `obj` into RTLIR representation
      assert False, 'unrecognized object {}!'.format( _obj )
    except AssertionError as e:
      msg = '' if e.args[0] is None else e.args[0]
      raise RTLIRConversionError( _obj, msg )



"""
========================================================================
ComponentLevel3_test.py
========================================================================

Author : Shunning Jiang
Date   : Dec 25, 2017
"""
from __future__ import absolute_import, division, print_function

from collections import deque

from pymtl3.datatypes import Bits8, Bits10, Bits32
from pymtl3.dsl.ComponentLevel3 import ComponentLevel3
from pymtl3.dsl.Connectable import InPort, OutPort, Wire
from pymtl3.dsl.ConstraintTypes import WR, U
from pymtl3.dsl.errors import InvalidConnectionError, MultiWriterError

from .sim_utils import simple_sim_pass


def _test_model( cls ):
  A = cls()
  A.elaborate()
  simple_sim_pass( A, 0x123 )

  T, time = 0, 20
  while not A.done() and T < time:
    A.tick()
    print(A.line_trace())
    T += 1

MUX_SEL_0 = 0
MUX_SEL_1 = 1

class TestSource( ComponentLevel3 ):

  def construct( s, Type, input_ ):
    assert type(input_) == list, "TestSrc only accepts a list of inputs!"

    s.Type = Type
    s.input_ = deque( input_ ) # deque.popleft() is faster
    s.out = OutPort( Type )

    @s.update
    def up_src():
      if not s.input_:
        s.out = s.Type()
      else:
        s.out = s.input_.popleft()

  def done( s ):
    return not s.input_

  def line_trace( s ):
    return "%s" % s.out

class TestSink( ComponentLevel3 ):

  def construct( s, Type, answer ):
    assert type(answer) == list, "TestSink only accepts a list of outputs!"

    s.answer = deque( answer )
    s.in_ = InPort( Type )

    @s.update
    def up_sink():
      if not s.answer:
        assert False, "Simulation has ended"
      else:
        ref = s.answer.popleft()
        ans = s.in_

        assert ref == ans or ref == "*", "Expect %s, get %s instead" % (ref, ans)

  def done( s ):
    return not s.answer

  def line_trace( s ):
    return "%s" % s.in_

class Mux( ComponentLevel3 ):

  def construct( s, Type, ninputs ):
    s.in_ = [ InPort( Type ) for _ in xrange(ninputs) ]
    s.sel = InPort( int if Type is int else mk_bits( clog2(ninputs) ) )
    s.out = OutPort( Type )

    @s.update
    def up_mux():
      s.out = s.in_[ s.sel ]

  def line_trace( s ):  pass

def test_connect_list_const_idx():

  class Top(ComponentLevel3):

    def construct( s ):

      s.src_in0 = TestSource( int, [4,3,2,1] )
      s.src_in1 = TestSource( int, [8,7,6,5] )
      s.src_sel = TestSource( int, [1,0,1,0] )
      s.sink    = TestSink  ( int, [8,3,6,1] )

      s.mux = Mux(int, 2)

      s.connect( s.mux.in_[MUX_SEL_0], s.src_in0.out )
      s.connect( s.mux.in_[MUX_SEL_1], s.src_in1.out )
      s.connect( s.mux.sel           , s.src_sel.out )
      s.connect( s.sink.in_          , s.mux.out     )

    def done( s ):
      return s.src_in0.done() and s.sink.done()

    def line_trace( s ):
      return " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_connect_list_idx_call():

  class Top(ComponentLevel3):

    def construct( s ):

      s.src_in0 = TestSource( int, [4,3,2,1] )
      s.src_in1 = TestSource( int, [8,7,6,5] )
      s.src_sel = TestSource( int, [1,0,1,0] )
      s.sink    = TestSink  ( int, [8,3,6,1] )

      s.mux = Mux(int, 2)(
        out = s.sink.in_,
        in_ = { MUX_SEL_0: s.src_in0.out, MUX_SEL_1: s.src_in1.out },
        sel = s.src_sel.out,
      )

    def done( s ):
      return s.src_in0.done() and s.sink.done()

    def line_trace( s ):
      return " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_connect_deep():

  class MuxWrap(ComponentLevel3):

    def construct( s ):
      s.in_ = [ InPort(int) for _ in xrange(2) ]
      s.sel = InPort(int)
      s.out = OutPort(int)

      s.mux = Mux(int, 2)(
        out = s.out,
        in_ = { 0: s.in_[0], 1: s.in_[1] },
        sel = s.sel,
      )

  class Top(ComponentLevel3):

    def construct( s ):

      s.src_in0 = TestSource( int, [4,3,2,1] )
      s.src_in1 = TestSource( int, [8,7,6,5] )
      s.src_sel = TestSource( int, [1,0,1,0] )
      s.sink    = TestSink  ( int, [8,3,6,1] )

      s.mux_wrap = MuxWrap()(
        out = s.sink.in_,
        in_ = { 0: s.src_in0.out, 1: s.src_in1.out },
        sel = s.src_sel.out,
      )

    def done( s ):
      return s.src_in0.done() and s.sink.done()

    def line_trace( s ):
      return " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_deep_connect():

  class MuxWrap3(ComponentLevel3):

    def construct( s ):
      s.in_ = [ InPort(int) for _ in xrange(2) ]
      s.sel = InPort(int)
      s.out = OutPort(int)

      s.mux1 = Mux(int, 2)(
        in_ = { 0: s.in_[0], 1: s.in_[1] },
        sel = s.sel,
      )
      s.mux2 = Mux(int, 2)(
        in_ = { 0: s.in_[0], 1: s.in_[1] },
        sel = s.sel,
      )
      s.mux3 = Mux(int, 2)(
        out = s.out,
        in_ = { 0: s.mux1.out, 1: s.mux2.out },
        sel = s.sel,
      )

  class Top(ComponentLevel3):

    def construct( s ):

      s.src_in0 = TestSource( int, [4,3,2,1] )
      s.src_in1 = TestSource( int, [8,7,6,5] )
      s.src_sel = TestSource( int, [1,0,1,0] )
      s.sink    = TestSink  ( int, [8,3,6,1] )

      s.mux_wrap = MuxWrap3()(
        out = s.sink.in_,
        in_ = { 0: s.src_in0.out, 1: s.src_in1.out },
        sel = s.src_sel.out,
      )

    def done( s ):
      return s.src_in0.done() and s.sink.done()

    def line_trace( s ):
      return " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_2d_array_vars_connect_impl():

  class Top(ComponentLevel3):

    def construct( s ):

      s.src  = TestSource( int, [2,1,0,2,1,0] )
      s.sink = TestSink  ( int, ["*",(5+6),(3+4),(1+2),
                                 (5+6),(3+4),(1+2)] )

      s.wire = [ [ Wire(int) for _ in xrange(2)] for _ in xrange(2) ]
      s.connect( s.wire[0][0], s.src.out )

      @s.update
      def up_from_src():
        s.wire[0][1] = s.src.out + 1

      s.reg = Wire(int)
      s.connect( s.wire[1][0], s.reg )

      @s.update_on_edge
      def up_reg():
        s.reg = s.wire[0][0] + s.wire[0][1]

      @s.update
      def upA():
        s.wire[1][1] = s.reg + 1

      @s.update
      def up_to_sink():
        s.sink.in_ = s.wire[1][0] + s.wire[1][1]

    def done( s ):
      return s.src.done() and s.sink.done()

    def line_trace( s ):
      return s.src.line_trace() + " >>> " + \
             str(s.wire)+" r0=%s" % s.reg + \
             " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_lots_of_fan_connect():

  class Top(ComponentLevel3):

    def construct( s ):

      s.src  = TestSource( int, [4,3,2,1,4,3,2,1] )
      s.sink = TestSink  ( int, ["*",(5+5+6+6),(4+4+5+5),(3+3+4+4),(2+2+3+3),
                                     (5+5+6+6),(4+4+5+5),(3+3+4+4),(2+2+3+3)] )

      s.wire0 = Wire(int)

      @s.update
      def up_from_src():
        s.wire0 = s.src.out + 1

      s.reg = Wire(int)

      @s.update_on_edge
      def up_reg():
        s.reg = s.wire0

      s.wire1 = Wire(int)
      s.wire2 = Wire(int)

      s.connect( s.wire1, s.reg )

      @s.update
      def upA():
        s.wire2 = s.reg + 1

      s.wire3 = Wire(int)
      s.wire4 = Wire(int)

      s.connect( s.wire3, s.wire1 )
      s.connect( s.wire4, s.wire1 )

      s.wire5 = Wire(int)
      s.wire6 = Wire(int)

      s.connect( s.wire5, s.wire2 )
      s.connect( s.wire6, s.wire2 )

      s.wire7 = Wire(int)
      s.wire8 = Wire(int)

      @s.update
      def upD():
        s.wire7 = s.wire3 + s.wire6
        s.wire8 = s.wire4 + s.wire5

      @s.update
      def up_to_sink():
        s.sink.in_ = s.wire7 + s.wire8

    def done( s ):
      return s.src.done() and s.sink.done()

    def line_trace( s ):
      return s.src.line_trace() + " >>> " + \
            "w0=%s > r0=%s" % (s.wire0,s.reg) + \
             " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_connect_plain():

  class Top(ComponentLevel3):

    def construct( s ):

      s.src  = TestSource( int, [4,3,2,1,4,3,2,1] )
      s.sink = TestSink  ( int, [5,4,3,2,5,4,3,2] )

      s.wire0 = Wire(int)

      @s.update
      def up_from_src():
        s.wire0 = s.src.out + 1

      s.connect( s.sink.in_, s.wire0 )

    def done( s ):
      return s.src.done() and s.sink.done()

    def line_trace( s ):
      return s.src.line_trace() + " >>> " + \
            "w0=%s" % (s.wire0) + \
             " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_2d_array_vars_connect():

  class Top(ComponentLevel3):

    def construct( s ):

      s.src  = TestSource( int, [2,1,0,2,1,0] )
      s.sink = TestSink  ( int, ["*",(5+6),(3+4),(1+2),
                                     (5+6),(3+4),(1+2)] )

      s.wire = [ [ Wire(int) for _ in xrange(2)] for _ in xrange(2) ]
      s.connect( s.wire[0][0], s.src.out )

      @s.update
      def up_from_src():
        s.wire[0][1] = s.src.out + 1

      s.reg = Wire(int)
      s.connect( s.wire[1][0], s.reg )

      @s.update
      def up_reg():
        s.reg = s.wire[0][0] + s.wire[0][1]

      for i in xrange(2):
        s.add_constraints(
          U(up_reg) < WR(s.wire[0][i]), # up_reg reads  s.wire[0][i]
        )

      @s.update
      def upA():
        s.wire[1][1] = s.reg + 1

      @s.update
      def up_to_sink():
        s.sink.in_ = s.wire[1][0] + s.wire[1][1]

    def done( s ):
      return s.src.done() and s.sink.done()

    def line_trace( s ):
      return s.src.line_trace() + " >>> " + \
             str(s.wire)+" r0=%s" % s.reg + \
             " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_connect_const_same_level():

  class Top(ComponentLevel3):

    def construct( s ):

      s.a = Wire(int)
      s.connect( s.a, 0 )

      @s.update
      def up_printa():
        print(s.a)

    def done( s ):
      return False

    def line_trace( s ):
      return ""

  _test_model( Top )

def test_connect_const_two_writer():

  class Top(ComponentLevel3):

    def construct( s ):

      s.a = Wire(int)
      s.connect( s.a, 0 )

      @s.update
      def up_printa():
        print(s.a)

      @s.update
      def up_writea():
        s.a = 123

    def done( s ):
      return False

    def line_trace( s ):
      return ""

  try:
    _test_model( Top )
  except MultiWriterError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown MultiWriterError.")

def test_connect_list_idx_call():

  class Top(ComponentLevel3):

    def construct( s ):

      s.src_in0 = TestSource( int, [4,3,2,1] )
      s.src_sel = TestSource( int, [1,0,1,0] )
      s.sink    = TestSink  ( int, [12,3,12,1] )

      s.mux = Mux(int, 2)(
        out = s.sink.in_,
        in_ = { MUX_SEL_0: s.src_in0.out },
        sel = s.src_sel.out,
      )
      s.connect( s.mux.in_[MUX_SEL_1], 12 )

    def done( s ):
      return s.src_in0.done() and s.sink.done()

    def line_trace( s ):
      return " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_connect_list_idx_const_in_call():

  class Top(ComponentLevel3):

    def construct( s ):

      s.src_in0 = TestSource( int, [4,3,2,1] )
      s.src_sel = TestSource( int, [1,0,1,0] )
      s.sink    = TestSink  ( int, [12,3,12,1] )

      s.mux = Mux(int, 2)(
        out = s.sink.in_,
        in_ = { MUX_SEL_0: s.src_in0.out, MUX_SEL_1: 12 },
        sel = s.src_sel.out,
      )

    def done( s ):
      return s.src_in0.done() and s.sink.done()

    def line_trace( s ):
      return " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_top_level_inport():

  class Top( ComponentLevel3 ):

    def construct( s ):

      s.a = InPort(Bits10)
      s.b = Wire(Bits32)
      s.connect( s.a, s.b[0:10] )

      @s.update
      def up():
        print(s.b[10:32])

    def done( s ):
      return False

    def line_trace( s ):
      return ""

  _test_model( Top )

def test_top_level_outport():

  class Top( ComponentLevel3 ):

    def construct( s ):

      s.a = OutPort(Bits10)
      s.b = Wire(Bits32)
      s.connect( s.a, s.b[9:19] )

      @s.update
      def up():
        s.b[0:10] = 1023

    def done( s ):
      return False

    def line_trace( s ):
      return str(int(s.a))

  A = Top()
  A.elaborate()
  simple_sim_pass( A )

  A.tick()
  trace = A.line_trace()
  print(" >>>", trace)
  assert trace == "1"

# This is a simplified test case from Peitian

def test_multiple_slices_are_net_writers():

  class A( ComponentLevel3 ):

    def construct( s ):
      s.in_  = InPort( Bits32 )
      s.out1 = OutPort( Bits8 )
      s.out2 = OutPort( Bits8 )

      s.connect( s.in_[0:8], s.out1[0:8] )
      s.connect( s.in_[0:4], s.out2[0:4] )

  a = A()
  a.elaborate()

# A similar case for struct

def test_multiple_fields_are_net_writers():

  class SomeMsg1( object ):
    def __init__( s, a=0, b=0 ):
      s.a = Bits8(a)
      s.b = Bits32(b)

  class SomeMsg2( object ):
    def __init__( s, a=0 ):
      s.c = Bits8(a)

  class A( ComponentLevel3 ):

    def construct( s ):
      s.in_  = InPort( SomeMsg1 )
      s.out1 = OutPort( SomeMsg2 )
      s.out2 = OutPort( SomeMsg2 )

      s.connect( s.in_.a, s.out1.c )
      s.connect( s.in_.b[0:8], s.out2.c )

  a = A()
  a.elaborate()
  simple_sim_pass( a )

  a.in_ = SomeMsg1(4, 5)
  a.tick()
  assert a.out1.c == 4

# The counterpart of the above test

def test_multiple_fields_are_assigned():

  class SomeMsg1( object ):
    def __init__( s, a=0, b=0 ):
      s.a = Bits8(a)
      s.b = Bits32(b)

  class SomeMsg2( object ):
    def __init__( s, a=0 ):
      s.c = Bits8(a)

  class A( ComponentLevel3 ):

    def construct( s ):
      s.in_  = InPort( SomeMsg1 )
      s.out1 = OutPort( SomeMsg2 )
      s.out2 = OutPort( SomeMsg2 )

      # s.connect( s.in_.a, s.out1.c )
      # s.connect( s.in_.b[0:8], s.out2.c )
      @s.update
      def up_pass():
        s.out1.c = s.in_.a
        s.out2.c = s.in_.b[0:8]

  a = A()
  a.elaborate()
  simple_sim_pass( a )

  a.in_ = SomeMsg1(4, 5)
  a.tick()
  assert a.out1.c == 4

def test_const_connect_struct_signal_to_int():

  class SomeMsg1( object ):
    def __init__( s, a=0, b=0 ):
      s.a = Bits8(a)
      s.b = Bits32(b)

    def __eq__( s, other ):
      return s.a == other.a and s.b == other.b

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.wire = Wire(SomeMsg1)
      s.connect( s.wire, 1 )

  try:
    x = Top()
    x.elaborate()
  except InvalidConnectionError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown InvalidConnectionError.")

def test_const_connect_struct_signal_to_Bits():

  class SomeMsg1( object ):
    def __init__( s, a=0, b=0 ):
      s.a = Bits8(a)
      s.b = Bits32(b)

    def __eq__( s, other ):
      return s.a == other.a and s.b == other.b

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.wire = Wire(SomeMsg1)
      s.connect( s.wire, Bits32(1) )

  try:
    x = Top()
    x.elaborate()
  except InvalidConnectionError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown InvalidConnectionError.")

def test_const_connect_Bits_signal_to_int():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.wire = Wire(Bits32)
      s.connect( s.wire, 1 )

  x = Top()
  x.elaborate()
  print(x._dsl.consts)
  assert len(x._dsl.consts) == 1

def test_const_connect_int_signal_to_int():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.wire = Wire(int)
      s.connect( s.wire, 1 )

  x = Top()
  x.elaborate()
  print(x._dsl.consts)
  assert len(x._dsl.consts) == 1

def test_const_connect_Bits_signal_to_Bits():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.wire = Wire(Bits32)
      s.connect( s.wire, Bits32(0) )

  x = Top()
  x.elaborate()
  print(x._dsl.consts)
  assert len(x._dsl.consts) == 1

def test_const_connect_Bits_signal_to_mismatch_Bits():

  class Top( ComponentLevel3 ):
    def construct( s ):
      s.wire = Wire(Bits32)
      s.connect( s.wire, Bits8(8) )

  try:
    x = Top()
    x.elaborate()
  except InvalidConnectionError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown InvalidConnectionError.")



from __future__ import absolute_import, division, print_function

from .datatypes import *
from .datatypes import _bitwidths
from .dsl.Component import Component
from .dsl.ComponentLevel5 import method_port
from .dsl.ComponentLevel6 import non_blocking
from .dsl.ComponentLevel7 import blocking
from .dsl.Connectable import (
    CalleePort,
    CallerPort,
    InPort,
    Interface,
    NonBlockingCalleeIfc,
    NonBlockingCallerIfc,
    OutPort,
    Wire,
)
from .dsl.ConstraintTypes import RD, WR, M, U
from .passes.PassGroups import DynamicSim, SimpleSim, SimulationPass

__version__ = "0.1.0"

__all__ = [
  'U','M','RD','WR',
  'Wire', 'InPort', 'OutPort', 'Interface', 'CallerPort', 'CalleePort',
  'method_port',
  'non_blocking', 'NonBlockingCalleeIfc', 'NonBlockingCallerIfc', 'blocking',

  'SimpleSim', 'DynamicSim', 'SimulationPass',
  'Component',

  'sext', 'zext', 'clog2', 'concat', 'reduce_and', 'reduce_or', 'reduce_xor',
  'mk_bits', 'Bits',
  'mk_bit_struct', 'BitStruct',
] + [ "Bits{}".format(x) for x in _bitwidths ] \
  + [ "b{}".format(x) for x in _bitwidths ]



"""
#=========================================================================
# BitStruct.py
#=========================================================================
# A basic BitStruct definition.
#
# Author : Yanghui Ou
#   Date : Apr 8, 2019
"""
from __future__ import absolute_import, division, print_function

from functools import reduce

import py

from .bits_import import *
from .helpers import concat


class BitStruct( object ):

  fields = []
  nbits = 0

  def to_bits( self ):
    bits_lst = []
    assert self.fields, \
        "fields of BitStruct {} are not specified!".format( self.__class__.__name__ )
    for name, Type in self.fields:
      if issubclass( Type, Bits ):
        bits_lst.append( Type( vars( self )[name] ) )
      elif issubclass( Type, BitStruct ):
        assert isinstance( vars( self )[name], BitStruct )
        bits_lst.append( vars( self )[name].to_bits() )
      else:
        raise AssertionError(
          "Type {} cannot be translated into bits!".format( Type )
        )
    return reduce( lambda x, y : concat( x, y ), bits_lst )

  # TODO: better error message.
  @classmethod
  def field_nbits( cls, field_name ):
    for fname, FieldType in cls.fields:
      if field_name == fname:
        return FieldType.nbits
    raise AssertionError(
      "{} does not have field {}!".format( cls.__name__, field_name )
    )

  # Default __str__ function

  def __str__( self ):
    trace = ""
    for name, _ in self.fields:
      trace += "{}:".format( vars( self )[name] )
    return trace[:-1]

  # Compare only fields
  # TODO: figure out how frequently user might directly compare the whole
  # struct as the comparison between structs can be slow. We assume this
  # would mainly happen in test sinks.
  def __eq__( self, other ):
    if type( self ) != type( other ):
      raise AssertionError(
        "Cannot compare {} against {}".format( type( self ), type( other ) )
      )
    # TODO: figure out whether we should just compare each fields
    # recursively so that it supports types that cannot be converted to bits?
    return self.to_bits() == other.to_bits()
  
  def __hash__( self ):
    return hash((self.__class__, self.to_bits()))

  def __ne__( self, other ):
    if type( self ) != type( other ):
      raise AssertionError(
        "Cannot compare {} against {}".format( type( self ), type( other ) )
      )
    return self.to_bits() != other.to_bits()

#-------------------------------------------------------------------------
# Dynamically generate bit struct
#-------------------------------------------------------------------------

_struct_dict = dict()
_fields_dict = dict()
_struct_tmpl = """
class {name}( BitStruct ):

  def __init__( s, {args_str} ):
{assign_str}

  def __hash__( s ):
    return hash(({hash_str}))

_struct_dict[\"{name}\"] = {name}
"""

def mk_bit_struct( name, fields, str_func=None ):

  if name in _struct_dict:
    if _fields_dict[name] == fields:
      return _struct_dict[name]
    else:
      raise AssertionError(
        "BitStruct {} has already been created!".format( name )
      )
  else:
    # Check for duplicate fields first.
    # TODO: better error message
    keys = [ fname for fname, _ in fields ]
    if any( keys.count(k) > 1 for k in keys):
      raise AssertionError(
        "Failed to create {} due to duplicate fields!".format( name )
      )

    args_str   = ""
    assign_str = ""
    hash_str = name + ", "
    nbits      = 0
    for field_name, FieldType in fields:
      nbits += FieldType.nbits
      args_str   += "{}={}(), ".format( field_name, FieldType.__name__ )
      assign_str += "    s.{} = {}\n".format( field_name, field_name )
      hash_str += "s.{}, ".format(field_name)
    args_str = args_str[:-2]

    tmpl = _struct_tmpl.format( **locals() )
    exec( py.code.Source( tmpl ).compile(), globals() )

    cls = _struct_dict[name]

    cls.fields = fields
    cls.nbits  = nbits
    if str_func is not None:
      cls.__str__ = str_func

    _fields_dict[name] = fields
    return _struct_dict[name]



#=========================================================================
# TranslationImport_stdlib_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jun 3, 2019
"""Test stdlib RTL components with SystemVerilog translation and import.

We reuse all the test cases from stdlib test files. To achieve this I
overwrite the original reference to the test function used in stdlib test
files and add my own test function (the `run_test`s). However we need
to make sure the orignal reference is not lost and is restored after
finishing each test (no matter it fails or passes).
"""

from __future__ import absolute_import, division, print_function

from pymtl3.datatypes import Bits1, mk_bits
from pymtl3.passes.rtlir.util.test_utility import do_test
from pymtl3.passes.sverilog import TranslationPass
from pymtl3.passes.sverilog.import_.ImportPass import ImportPass
from pymtl3.stdlib.rtl.arbiters_test import test_rr_arb_4 as _rr_arb_4
from pymtl3.stdlib.rtl.arbiters_test import test_rr_arb_en_4 as _rr_arb_en_4
from pymtl3.stdlib.rtl.Crossbar_test import test_crossbar3 as _crossbar3
from pymtl3.stdlib.rtl.Encoder_test import test_encoder_5_directed as _encoder5
from pymtl3.stdlib.rtl.valrdy_queues_test import test_2entry_normal_Bits as _n2
from pymtl3.stdlib.rtl.valrdy_queues_test import test_3entry_normal_Bits as _n3
from pymtl3.stdlib.rtl.valrdy_queues_test import test_bypass_Bits as _bypass_Bits
from pymtl3.stdlib.rtl.valrdy_queues_test import test_bypass_int as _bypass_int
from pymtl3.stdlib.rtl.valrdy_queues_test import test_normal_Bits as _normal_Bits
from pymtl3.stdlib.rtl.valrdy_queues_test import test_normal_int as _normal_int
from pymtl3.stdlib.rtl.valrdy_queues_test import test_pipe_Bits as _pipe_Bits
from pymtl3.stdlib.rtl.valrdy_queues_test import test_pipe_int as _pipe_int
from pymtl3.stdlib.test import TestVectorSimulator


def local_do_test( _m ):
  try:
    _m.elaborate()
    # Mark component `_m` as to be translated and imported
    _m.sverilog_translate = True
    _m.sverilog_import = True
    _m.apply( TranslationPass() )
    # We are importing the top component and therefore should expect a new top
    # as the return value.
    m = ImportPass()( _m )
    sim = TestVectorSimulator( m, _m._test_vectors, _m._tv_in, _m._tv_out )
    sim.run_test()
  finally:
    try:
      # Explicitly finalize imported component to avoid shared lib name aliasing
      m.finalize()
    except UnboundLocalError:
      # This test fails due to translation errors
      pass

def test_arbiter_rr_arb_4( do_test ):
  # Customized test function
  def run_test( cls, args, test_vectors ):
    m = cls( *args )
    BitsN = mk_bits( args[0] )
    def tv_in( model, test_vector ):
      model.reqs = BitsN(test_vector[0])
    def tv_out( model, test_vector ):
      assert model.grants == BitsN(test_vector[1])
    m._test_vectors = test_vectors
    m._tv_in, m._tv_out = tv_in, tv_out
    do_test( m )

  # Swap in the customized test method
  test_func = _rr_arb_4
  _run_test = test_func.__globals__['run_test']
  test_func.__globals__['run_test'] = run_test
  try:
    test_func()
  finally:
    # Restore the original method regardless of the test result
    test_func.__globals__['run_test'] = _run_test

def test_arbiter_rr_arb_en_4( do_test ):
  def run_test( cls, args, test_vectors ):
    m = cls( *args )
    BitsN = mk_bits( args[0] )
    def tv_in( model, test_vector ):
      model.en   = Bits1( test_vector[0] )
      model.reqs = BitsN( test_vector[1] )
    def tv_out( model, test_vector ):
      assert model.grants == BitsN(test_vector[2])
    m._test_vectors = test_vectors
    m._tv_in, m._tv_out = tv_in, tv_out
    do_test( m )

  test_func = _rr_arb_en_4
  _run_test = test_func.__globals__['run_en_test']
  test_func.__globals__['run_en_test'] = run_test
  try:
    test_func( None, None )
  finally:
    test_func.__globals__['run_en_test'] = _run_test

def test_crossbar3( do_test ):
  def run_test( cls, args, test_vectors ):
    m = cls( *args )
    T = args[1]
    def tv_in( model, test_vector ):
      n = len( model.in_ )
      for i in range(n):
        model.in_[i] = T(test_vector[i])
        model.sel[i] = T(test_vector[n+i])
    def tv_out( model, test_vector ):
      n = len( model.in_ )
      for i in range(n):
        assert model.out[i] == T(test_vector[n*2+i])
    m._test_vectors = test_vectors
    m._tv_in, m._tv_out = tv_in, tv_out
    do_test( m )

  test_func = _crossbar3
  _run_test = test_func.__globals__['run_test_crossbar']
  test_func.__globals__['run_test_crossbar'] = run_test
  try:
    test_func()
  finally:
    test_func.__globals__['run_test_crossbar'] = _run_test

def test_encoder_5_directed( do_test ):
  def run_test( m, test_vectors ):
    def tv_in( model, test_vector ):
      model.in_ = test_vector[0]
    def tv_out( model, test_vector ):
      assert model.out == test_vector[1]
    m._test_vectors = test_vectors
    m._tv_in, m._tv_out = tv_in, tv_out
    do_test( m )

  test_func = _encoder5
  _run_test = test_func.__globals__['run_test']
  test_func.__globals__['run_test'] = run_test
  try:
    test_func()
  finally:
    test_func.__globals__['run_test'] = _run_test

#-------------------------------------------------------------------------
# Queue tests
#-------------------------------------------------------------------------

def _queue_test_tv_in( model, tv ):
  model.enq.val = tv[0]
  model.enq.msg = tv[2]
  model.deq.rdy = tv[4]

def _queue_test_tv_out( model, tv ):
  if tv[1] != '?': assert model.enq.rdy == tv[1]
  if tv[3] != '?': assert model.deq.val == tv[3]
  if tv[5] != '?': assert model.deq.msg == tv[5]

def test_bypass_Bits( do_test ):
  def _run_queue_test( m, test_vectors ):
    m._test_vectors = test_vectors
    m._tv_in, m._tv_out = _queue_test_tv_in, _queue_test_tv_out
    do_test( m )
  test_func = _bypass_Bits
  _run_test = test_func.__globals__['run_test_queue']
  test_func.__globals__['run_test_queue'] = _run_queue_test
  try:
    test_func()
  finally:
    test_func.__globals__['run_test_queue'] = _run_test

def test_pipe_Bits( do_test ):
  def _run_queue_test( m, test_vectors ):
    m._test_vectors = test_vectors
    m._tv_in, m._tv_out = _queue_test_tv_in, _queue_test_tv_out
    do_test( m )
  test_func = _pipe_Bits
  _run_test = test_func.__globals__['run_test_queue']
  test_func.__globals__['run_test_queue'] = _run_queue_test
  try:
    test_func()
  finally:
    test_func.__globals__['run_test_queue'] = _run_test

def test_normal_Bits( do_test ):
  def _run_queue_test( m, test_vectors ):
    m._test_vectors = test_vectors
    m._tv_in, m._tv_out = _queue_test_tv_in, _queue_test_tv_out
    do_test( m )
  test_func = _normal_Bits
  _run_test = test_func.__globals__['run_test_queue']
  test_func.__globals__['run_test_queue'] = _run_queue_test
  try:
    test_func()
  finally:
    test_func.__globals__['run_test_queue'] = _run_test

def test_2entry_normal_Bits( do_test ):
  def _run_queue_test( m, test_vectors ):
    m._test_vectors = test_vectors
    m._tv_in, m._tv_out = _queue_test_tv_in, _queue_test_tv_out
    do_test( m )
  test_func = _n2
  _run_test = test_func.__globals__['run_test_queue']
  test_func.__globals__['run_test_queue'] = _run_queue_test
  try:
    test_func()
  finally:
    test_func.__globals__['run_test_queue'] = _run_test

def test_3entry_normal_Bits( do_test ):
  def _run_queue_test( m, test_vectors ):
    m._test_vectors = test_vectors
    m._tv_in, m._tv_out = _queue_test_tv_in, _queue_test_tv_out
    do_test( m )
  test_func = _n3
  _run_test = test_func.__globals__['run_test_queue']
  test_func.__globals__['run_test_queue'] = _run_queue_test
  try:
    test_func()
  finally:
    test_func.__globals__['run_test_queue'] = _run_test



"""
========================================================================
VcdGenerationPass.py
========================================================================

Author : Shunning Jiang
Date   : May 29, 2019
"""
from __future__ import absolute_import, division, print_function

import time
from collections import defaultdict
from copy import deepcopy

import py

from pymtl3.dsl import Const
from pymtl3.passes.BasePass import BasePass, PassMetadata

from .errors import PassOrderError


class VcdGenerationPass( BasePass ):

  def __call__( self, top ):

    # Check for dum_vcd flag
    if not hasattr( top, "dump_vcd" ):
      return
    if not top.dump_vcd:
      return

    if not hasattr( top._sched, "schedule" ):
      raise PassOrderError( "schedule" )

    if hasattr( top, "_cl_trace" ):
      schedule = top._cl_trace.schedule
    else:
      schedule = top._sched.schedule

    top._vcd = PassMetadata()

    schedule.append( self.make_vcd_func( top, top._vcd ) )

  def make_vcd_func( self, top, vcdmeta ):

    if hasattr( top, "vcd_file_name" ):
      vcdmeta.vcd_file_name = str(top.vcd_file_name) + ".vcd"
    else:
      vcdmeta.vcd_file_name = str(top.__class__) + ".vcd"
    vcdmeta.vcd_file = open( vcdmeta.vcd_file_name, "w" )

    # Get vcd timescale

    try:                    vcd_timescale = top.vcd_timescale
    except AttributeError:  vcd_timescale = "10ps"

    # Print vcd header

    print( "$date\n    {}\n$end\n$version\n    PyMTL 3 (Mamba)\n$end\n"
           "$timescale {}\n$end\n".format( time.asctime(), vcd_timescale ),
           file=vcdmeta.vcd_file )

    # Utility generator to create new symbols for each VCD signal.
    # Code inspired by MyHDL 0.7.
    # Shunning: I just reuse it from pymtl v2

    def _gen_vcd_symbol():

      # Generate a string containing all valid vcd symbol characters
      _codechars = ''.join([chr(i) for i in range(33, 127)])
      _mod       = len(_codechars)

      # Generator logic
      n = 0
      while True:
        q, r = divmod(n, _mod)
        code = _codechars[r]
        while q > 0:
          q, r = divmod(q, _mod)
          code = _codechars[r] + code
        yield code
        n += 1

    vcd_symbols = _gen_vcd_symbol()

    # Preprocess some metadata

    component_signals = defaultdict(set)

    all_components = set()

    # We only collect non-sliced leaf signals
    # TODO only collect leaf signals and for nested structs
    for x in top._dsl.all_signals:
      for y in x.get_leaf_signals():
        host = y.get_host_component()
        component_signals[ host ].add(y)

    # We pre-process all nets in order to remove all sliced wires because
    # they belong to a top level wire and we count that wire

    trimmed_value_nets = []
    vcdmeta.clock_net_idx = None

    # FIXME handle the case where the top level signal is in a value net
    for writer, net in top.get_all_value_nets():
      new_net = []
      for x in net:
        if not isinstance(x, Const) and not x.is_sliced_signal():
          new_net.append( x )
          if repr(x) == "s.clk":
            # Hardcode clock net because it needs to go up and down
            assert vcdmeta.clock_net_idx is None
            vcdmeta.clock_net_idx = len(trimmed_value_nets)

      if new_net:
        trimmed_value_nets.append( new_net )

    # Generate symbol for existing nets

    net_symbol_mapping = [ next(vcd_symbols) for x in trimmed_value_nets ]
    signal_net_mapping = {}

    for i in range(len(trimmed_value_nets)):
      for x in trimmed_value_nets[i]:
        signal_net_mapping[x] = i

    # Inner utility function to perform recursive descent of the model.
    # Shunning: I mostly follow v2's implementation

    # Vcd file takes a(0) instead of a[0]
    def vcd_mangle_name( name ):
      return name.replace('[','(').replace(']',')')

    def recurse_models( m, level ):

      # Special case the top level "s" to "top"

      my_name = m.get_field_name()
      if my_name == "s":
        my_name = "top"

      # Create a new scope for this module
      print( "{}$scope module {} $end".format( "    "*level,
              vcd_mangle_name( my_name ) ),
              file=vcdmeta.vcd_file )

      m_name = repr(m)

      # Define all signals for this model.
      for signal in component_signals[m]:

        # Multiple signals may be collapsed into a single net in the
        # simulator if they are connected. Generate new vcd symbols per
        # net, not per signal as an optimization.

        if signal in signal_net_mapping:
          net_id = signal_net_mapping[signal]
          symbol = net_symbol_mapping[net_id]
        else:
          # We treat this as a new net

          # Check if it's clock. Hardcode clock net
          if repr(signal) == "s.clk":
            assert vcdmeta.clock_net_idx is None
            vcdmeta.clock_net_idx = len(trimmed_value_nets)

          trimmed_value_nets.append( [ signal ] )
          signal_net_mapping[signal] = len(signal_net_mapping)
          symbol = next(vcd_symbols)
          net_symbol_mapping.append( symbol )

        # This signal can be a part of an interface so we have to
        # "subtract" host component's name from signal's full name
        # to get the actual name like enq.rdy
        # TODO struct
        signal_name = vcd_mangle_name( repr(signal)[ len(m_name)+1: ] )
        print( "{}$var {type} {nbits} {symbol} {name} $end".format( "    "*(level+1),
                type='reg', nbits=signal._dsl.Type.nbits, symbol=symbol, name= signal_name),
              file=vcdmeta.vcd_file )

      # Recursively visit all submodels.
      for child in m.get_child_components():
        recurse_models( child, level+1 )

      print( "{}$upscope $end".format("    "*level),
              file=vcdmeta.vcd_file )

    # Begin recursive descent from the top-level model.
    recurse_models( top, 0 )

    # Once all models and their signals have been defined, end the
    # definition section of the vcd and print the initial values of all
    # nets in the design.
    print( "$enddefinitions $end\n", file=vcdmeta.vcd_file )

    for i, net in enumerate(trimmed_value_nets):
      print( "b{value} {symbol}".format(
          value=net[0]._dsl.Type().bin(), symbol=net_symbol_mapping[i],
      ), file=vcdmeta.vcd_file )

      # Set this to be the last cycle value
      setattr( vcdmeta, "last_{}".format(i), net[0]._dsl.Type().bin() )

    # Now we create per-cycle signal value collect functions

    vcdmeta.sim_ncycles = 0

    # Flip clock for the first cycle
    print( '\n#0\nb0b1 {}\n'.format( net_symbol_mapping[ vcdmeta.clock_net_idx ] ),
           file=vcdmeta.vcd_file )

    dump_vcd_per_signal = """
    if vcdmeta.last_{0} != {1}:
      try:
        value_str = {1}.bin()
      except AttributeError as e:
        raise AttributeError( '{{}}\\n - {1} becomes another type. Please check your code.'.format(e) )
      print( 'b{{}} {2}'.format( value_str ), file=vcdmeta.vcd_file )
      vcdmeta.last_{0} = deepcopy({1})"""

    # TODO type check

    # Concatenate the strings for all signals

    # Give all ' and " characters a preceding backslash for .format
    for i, x in enumerate(net_symbol_mapping):
      net_symbol_mapping[i] = x.replace('\\', '\\\\').replace('\'','\\\'').replace('\"','\\\"')

    vcd_srcs = []
    for i, net in enumerate( trimmed_value_nets ):
      if i != vcdmeta.clock_net_idx:
        symbol = net_symbol_mapping[i]
        vcd_srcs.append( dump_vcd_per_signal.format( i, net[0], symbol ) )

    deepcopy # I have to do this to circumvent the tools

    src =  """
def dump_vcd():

  try:
    # Type check
    {1}
    # Dump VCD
    {2}
  except Exception:
    raise

  # Flop clock at the end of cycle
  print( '\\n#{{}}\\nb0b0 {0}'.format(100*vcdmeta.sim_ncycles+50 ),
        file=vcdmeta.vcd_file )
  # Flip clock of the next cycle
  print( '#{{}}\\nb0b1 {0}\\n'.format( 100*vcdmeta.sim_ncycles+100 ), file=vcdmeta.vcd_file )
  vcdmeta.sim_ncycles += 1
""".format( net_symbol_mapping[ vcdmeta.clock_net_idx ], "", "".join(vcd_srcs) )

    s = top
    exec(compile( src, filename="vcd_generation", mode="exec"), globals().update(locals()))
    return dump_vcd



"""
========================================================================
UnrollTickPass.py
========================================================================
Generate an unrolled tick function

Author : Shunning Jiang
Date   : Dec 26, 2018
"""


from __future__ import absolute_import, division, print_function

import py

from pymtl3.passes.BasePass import BasePass
from pymtl3.passes.errors import PassOrderError


class UnrollTickPass( BasePass ):

  def __call__( self, top ):
    if not hasattr( top._sched, "schedule" ):
      raise PassOrderError( "schedule" )

    schedule = top._sched.schedule

    # Berkin IlBeyi's recipe
    strs = map( "  update_blk{}() # {}".format, xrange( len(schedule) ), \
                                              [ x.__name__ for x in schedule ] )
    gen_tick_src = """
        {}
        def tick_unroll():
          # The code below does the actual calling of update blocks.
          {}""".format( "; ".join( map(
                        "update_blk{0} = schedule[{0}]".format,
                        xrange( len( schedule ) ) ) ),
                        "\n          ".join( strs ) )

    exec(py.code.Source( gen_tick_src ).compile(), locals())

    top.tick = tick_unroll



"""
========================================================================
ComponentLevel2_test.py
========================================================================

Author : Shunning Jiang
Date   : Nov 3, 2018
"""
from __future__ import absolute_import, division, print_function

from collections import deque

from pymtl3.datatypes import Bits32
from pymtl3.dsl.ComponentLevel2 import ComponentLevel2
from pymtl3.dsl.Connectable import InPort, OutPort, Wire
from pymtl3.dsl.ConstraintTypes import RD, WR, U
from pymtl3.dsl.errors import (
    InvalidConstraintError,
    InvalidFuncCallError,
    UpblkCyclicError,
    VarNotDeclaredError,
)

from .sim_utils import simple_sim_pass


def _test_model( cls ):
  A = cls()
  A.elaborate()
  simple_sim_pass( A, 0x123 )

  T, time = 0, 20
  while not A.done() and T < time:
    A.tick()
    print(A.line_trace())
    T += 1

class TestSource( ComponentLevel2 ):

  def construct( s, input_ ):
    assert type(input_) == list, "TestSrc only accepts a list of inputs!"

    s.input_ = deque( input_ ) # deque.popleft() is faster
    s.out = OutPort(int)

    @s.update
    def up_src():
      if not s.input_:
        s.out = 0
      else:
        s.out = s.input_.popleft()

  def done( s ):
    return not s.input_

  def line_trace( s ):
    return "%s" % s.out

class TestSink( ComponentLevel2 ):

  def construct( s, answer ):
    assert type(answer) == list, "TestSink only accepts a list of outputs!"

    s.answer = deque( answer )
    s.in_ = InPort(int)

    @s.update
    def up_sink():
      if not s.answer:
        assert False, "Simulation has ended"
      else:
        ref = s.answer.popleft()
        ans = s.in_

        assert ref == ans or ref == "*", "Expect %s, get %s instead" % (ref, ans)

  def done( s ):
    return not s.answer

  def line_trace( s ):
    return "%s" % s.in_

def test_simple():

  class Top(ComponentLevel2):

    def construct( s ):
      s.a = Wire(int)
      s.b = Wire(int)

      @s.update
      def upA():
        s.a = s.b + 1

      @s.update
      def upB():
        s.b = s.b + 1

    def done( s ):
      return True

  _test_model( Top )

def test_cyclic_impl_dependency():

  class Top(ComponentLevel2):

    def construct( s ):
      s.a = Wire(int)
      s.b = Wire(int)

      @s.update
      def upA():
        s.a = s.b

      @s.update
      def upB():
        s.b = s.a

    def done( s ):
      return True

  try:
    _test_model( Top )
  except UpblkCyclicError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown UpblkCyclicError.")

def test_invalid_dependency():

  class Top(ComponentLevel2):

    def construct( s ):

      s.a = Wire(int)
      s.b = Wire(int)

      s.add_constraints(
        WR(s.a) < RD(s.b),
      )

  try:
    _test_model( Top )
  except InvalidConstraintError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return

  raise Exception("Should've thrown InvalidConstraintError.")

def test_variable_not_declared():

  class SomeMsg( object ):

    def __init__( s, a=0, b=0 ):
      s.a = int( a )
      s.b = Bits32( b )

  class A(ComponentLevel2):
    def construct( s ):
      s.a = Wire( SomeMsg )
      s.b = Wire( int )

      @s.update
      def upA():
        s.a.a.zzz = s.b + 1

      @s.update
      def upB():
        s.b = s.b + 1

  class Top(ComponentLevel2):

    def construct( s ):
      s.x = A()

    def done( s ):
      return True

  try:
    _test_model( Top )
  except VarNotDeclaredError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown VarNotDeclaredError.")

def test_2d_array_vars():

  class Top(ComponentLevel2):

    def construct( s ):

      s.src  = TestSource( [2,1,0,2,1,0] )
      s.sink = TestSink  ( ["*",(5+6),(3+4),(1+2),
                                (5+6),(3+4),(1+2)] )

      s.wire = [ [ Wire(int) for _ in xrange(2)] for _ in xrange(2) ]

      @s.update
      def up_from_src():
        s.wire[0][0] = s.src.out
        s.wire[0][1] = s.src.out + 1

      s.reg = Wire(int)

      @s.update
      def up_reg():
        s.reg = s.wire[0][0] + s.wire[0][1]

      s.add_constraints(
        U(up_reg) < RD(s.reg), # up_reg writes s.reg
      )

      @s.update
      def upA():
        for i in xrange(2):
          s.wire[1][i] = s.reg + i

      for i in xrange(2):
        s.add_constraints(
          U(up_reg) < WR(s.wire[0][i]), # up_reg reads  s.wire[0][i]
        )

      @s.update
      def up_to_sink():
        s.sink.in_ = s.wire[1][0] + s.wire[1][1]

      up_sink = s.sink.get_update_block("up_sink")

      s.add_constraints(
        U(upA)        < U(up_to_sink),
        U(up_to_sink) < U(up_sink),
      )

    def done( s ):
      return s.src.done() and s.sink.done()

    def line_trace( s ):
      return s.src.line_trace() + " >>> " + \
             str(s.wire)+"r0=%s" % s.reg + \
             " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_wire_up_constraint():

  class Top(ComponentLevel2):

    def construct( s ):

      s.src  = TestSource( [4,3,2,1,4,3,2,1] )
      s.sink = TestSink  ( [5,4,3,2,5,4,3,2] )

      @s.update
      def up_from_src():
        s.sink.in_ = s.src.out + 1

      s.add_constraints(
        U(up_from_src) < RD(s.sink.in_),
      )

    def done( s ):
      return s.src.done() and s.sink.done()

    def line_trace( s ):
      return s.src.line_trace() + " >>> " + \
             " >>> " + s.sink.line_trace()

  _test_model( Top )

# write two disjoint slices
def test_write_two_disjoint_slices():

  class Top(ComponentLevel2):
    def construct( s ):
      s.A  = Wire( Bits32 )

      @s.update
      def up_wr_0_16():
        s.A[0:16] = Bits16( 0xff )

      @s.update
      def up_wr_16_30():
        s.A[16:30] = Bits16( 0xff )

      @s.update
      def up_rd_12_30():
        assert s.A[12:30] == 0xff0

    def done( s ):
      return True

  _test_model( Top )

# WR A.b - RD A
def test_wr_A_b_rd_A_impl():


  class SomeMsg( object ):

    def __init__( s, a=0, b=0 ):
      s.a = int(a)
      s.b = Bits32(b)

  class Top(ComponentLevel2):
    def construct( s ):
      s.A  = Wire( SomeMsg )

      @s.update
      def up_wr_A_b():
        s.A.b = 123

      @s.update
      def up_rd_A():
        z = s.A

    def done( s ):
      return True

  _test_model( Top )

def test_add_loopback():

  class Top(ComponentLevel2):

    def construct( s ):

      s.src  = TestSource( [4,3,2,1] )
      s.sink = TestSink  ( ["*",(4+1),(3+1)+(4+1),(2+1)+(3+1)+(4+1),(1+1)+(2+1)+(3+1)+(4+1)] )

      s.wire0 = Wire(int)
      s.wire1 = Wire(int)

      @s.update
      def up_from_src():
        s.wire0 = s.src.out + 1

      s.reg0 = Wire(int)

      @s.update
      def upA():
        s.reg0 = s.wire0 + s.wire1

      @s.update
      def up_to_sink_and_loop_back():
        s.sink.in_ = s.reg0
        s.wire1 = s.reg0

      s.add_constraints(
        U(upA) < WR(s.wire1),
        U(upA) < WR(s.wire0),
        U(upA) < RD(s.reg0), # also implicit
      )

    def done( s ):
      return s.src.done() and s.sink.done()

    def line_trace( s ):
      return s.src.line_trace() + " >>> " + \
            "w0=%s > r0=%s > w1=%s" % (s.wire0,s.reg0,s.wire1) + \
             " >>> " + s.sink.line_trace()
  _test_model( Top )

def test_add_loopback_on_edge():

  class Top(ComponentLevel2):

    def construct( s ):

      s.src  = TestSource( [4,3,2,1] )
      s.sink = TestSink  ( ["*",(4+1),(3+1)+(4+1),(2+1)+(3+1)+(4+1),(1+1)+(2+1)+(3+1)+(4+1)] )

      s.wire0 = Wire(int)
      s.wire1 = Wire(int)

      @s.update
      def up_from_src():
        s.wire0 = s.src.out + 1

      s.reg0 = Wire(int)

      # UPDATE ON EDGE!
      @s.update_on_edge
      def upA():
        s.reg0 = s.wire0 + s.wire1

      @s.update
      def up_to_sink_and_loop_back():
        s.sink.in_ = s.reg0
        s.wire1 = s.reg0

    def done( s ):
      return s.src.done() and s.sink.done()

    def line_trace( s ):
      return s.src.line_trace() + " >>> " + \
            "w0=%s > r0=%s > w1=%s" % (s.wire0,s.reg0,s.wire1) + \
             " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_2d_array_vars_impl():

  class Top(ComponentLevel2):

    def construct( s ):

      s.src  = TestSource( [2,1,0,2,1,0] )
      s.sink = TestSink  ( ["*",(5+6),(3+4),(1+2),
                                (5+6),(3+4),(1+2)] )

      s.wire = [ [ Wire(int) for _ in xrange(2)] for _ in xrange(2) ]

      @s.update
      def up_from_src():
        s.wire[0][0] = s.src.out
        s.wire[0][1] = s.src.out + 1

      s.reg = Wire(int)

      @s.update_on_edge
      def up_reg():
        s.reg = s.wire[0][0] + s.wire[0][1]

      @s.update
      def upA():
        for i in xrange(2):
          s.wire[1][i] = s.reg + i

      @s.update
      def up_to_sink():
        s.sink.in_ = s.wire[1][0] + s.wire[1][1]

    def done( s ):
      return s.src.done() and s.sink.done()

    def line_trace( s ):
      return s.src.line_trace() + " >>> " + \
             str(s.wire)+"r0=%s" % s.reg + \
             " >>> " + s.sink.line_trace()

  _test_model( Top )

def test_simple_func_impl():

  class Top(ComponentLevel2):

    def construct( s ):
      s.a = Wire(int)
      s.b = Wire(int)

      s.counter_assign = Wire(int)
      s.counter_read   = Wire(int)

      @s.func
      def assignb( b ):
        s.b = b + (s.counter_assign == -1) # never -1

      @s.update
      def up_write():
        if s.counter_assign & 1:
          assign( 1, 2 )
        else:
          assign( 10, 20 )
        s.counter_assign += 1

      @s.update
      def up_read():
        if s.counter_read & 1:
          assert s.a == 1 and s.b == min(100,2)
        else:
          assert s.a == 10 and s.b == 20
        s.counter_read += 1

      # The order doesn't matter. As a result, funcs should be processed
      # after construction time
      @s.func
      def assign( a, b ):
        s.a = a + (s.counter_assign == -1)
        assignb( b )

    def done( s ):
      return False

    def line_trace( s ):
      return "{} {}".format( s.a, s.b )

  _test_model( Top )

def test_func_cyclic_invalid():

  class Top(ComponentLevel2):

    def construct( s ):
      s.a = Wire(int)
      s.b = Wire(int)
      s.c = Wire(int)

      @s.func
      def assignc( a, b ):
        s.c = a
        assign( a, b )

      @s.func
      def assignb( a, b ):
        s.b = b
        assignc( a, b )

      @s.func
      def assign( a, b ):
        s.a = b
        assignb( b, a )

      @s.update
      def up_write():
        assign( 1, 2 )

    def done( s ):
      return False

    def line_trace( s ):
      return "{} {}".format( s.a, s.b )

  try:
    _test_model( Top )
  except InvalidFuncCallError as e:
    print("{} is thrown\n{}".format( e.__class__.__name__, e ))
    return
  raise Exception("Should've thrown InvalidFuncCallError.")



#=========================================================================
# utility.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 27, 2019
"""Provide helper methods that might be useful to sverilog passes."""

from __future__ import absolute_import, division, print_function


def make_indent( src, nindent ):
  """Add nindent indention to every line in src."""
  indent = '  '
  for idx, s in enumerate( src ):
    src[ idx ] = nindent * indent + s

def get_component_unique_name( c_rtype ):

  def get_string( obj ):
    """Return the string that identifies `obj`"""
    if isinstance(obj, type): return obj.__name__
    return str( obj )

  comp_name = c_rtype.get_name()
  comp_params = c_rtype.get_params()
  assert comp_name
  for arg_name, arg_value in comp_params:
    assert arg_name != ''
    comp_name += '__' + arg_name + '_' + get_string(arg_value)
  return comp_name
